from __future__ import annotations
from typing import TYPE_CHECKING, Optional

from instaui.internal.codegen.components.codegen_components import CodegenComponents


if TYPE_CHECKING:
    from instaui.internal.backend.binding import BackendBindingRegistryBase


class CodegenRootContext:
    def __init__(
        self,
        *,
        binding_registry: BackendBindingRegistryBase,
        components: Optional[CodegenComponents] = None,
    ):
        self.binding_registry = binding_registry
        self.components = components or CodegenComponents.default()



import base64
from pathlib import Path
from urllib.parse import quote
from enum import Enum
import mimetypes
from .resolver import AssetResolver


class PREFIX(str, Enum):
    JS = "data:text/javascript;charset=utf-8"
    CSS = "data:text/css;charset=utf-8"
    ICON = "data:image/x-icon;base64"


PREFIX_MAP = {
    "js": PREFIX.JS.value,
    "mjs": PREFIX.JS.value,
    "css": PREFIX.CSS.value,
    "ico": PREFIX.ICON.value,
}


class ZeroAssetResolver(AssetResolver):
    def resolve_asset(self, path: Path) -> str:
        return _normalize_path_to_dataurl(path)


def _normalize_path_to_dataurl(path: Path):
    prefix, ext = _get_prefix(path)

    if ext == "ico":
        return _normalize_path_to_base64_url(path, prefix)

    content = path.read_text(encoding="utf-8")
    return f"{prefix},{quote(content)}"


def _normalize_path_to_base64_url(path: Path, prefix: str):
    return f"{prefix},{base64.b64encode(path.read_bytes()).decode('utf-8')}"


def _get_prefix(path: Path):
    ext = path.suffix.lstrip(".")
    prefix = PREFIX_MAP.get(ext, None)
    if not prefix:
        mime, _ = mimetypes.guess_type(path)
        if not mime:
            raise RuntimeError(f"Unknown mime type: {path}")

        prefix = f"data:{mime};base64"

    return prefix, ext



from typing import Optional
from instaui.internal.pipeline.utils import html_utils
from instaui.systems.dataclass_system import dataclass, field


@dataclass()
class HtmlRenderContext:
    app_var: str
    is_debug: bool = False
    lang: str = "en"
    title: str = ""
    favicon_url: str = ""
    prefix: str = ""

    css_links: list[str] = field(default_factory=list)
    style_tags: list[tuple[Optional[str], str]] = field(default_factory=list)
    """(group_id, content)"""

    js_links: list[str] = field(default_factory=list)

    script_tags: list[tuple[Optional[dict], str]] = field(default_factory=list)
    """(attrs_dict, content)"""
    entry_script_type: str = "module"
    entry_code: str = ""

    mount_id: str = "app"

    import_map: Optional[dict[str, str]] = None

    icons_svg_code: Optional[str] = None

    def render_attrs(self, attrs: Optional[dict]) -> str:
        return html_utils.render_attrs(attrs)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from instaui.internal.ast.symbol import SymbolId
from instaui.internal.pipeline.ast_build.expr_watch_parser import ExprWatchParser
from instaui.internal.pipeline.ast_build.js_watch_parser import JsWatchParser
from instaui.internal.pipeline.ast_build.render_builder import RenderBuilder
from instaui.internal.pipeline.ast_build.variable_decl_builder import (
    VariableDeclBuilder,
)

from instaui.internal.pipeline.ast_build.web_watch_task import WebWatchTaskParser
from instaui.internal.ui._scope import Scope
from instaui.internal.ast import core as ast_core
from instaui.internal.pipeline.ast_build.context import AstBuilderContext
from instaui.internal.ui.vfor_item import VForIndex, VForItem, VForItemKey

if TYPE_CHECKING:
    from instaui.internal.runtime.runtime_ctx import RuntimeContext
    from instaui.internal.runtime.render_ctx import RenderContext


class ComponentBuilder:
    def __init__(
        self,
        runtime_ctx: RuntimeContext,
        render_ctx: RenderContext,
        builder_ctx: AstBuilderContext,
    ) -> None:
        self.runtime_ctx = runtime_ctx
        self.ctx = builder_ctx
        self._variable_builder = VariableDeclBuilder(self.ctx)
        self._render_builder = RenderBuilder(self.ctx, self)
        self._web_watch_parser = WebWatchTaskParser(runtime_ctx, render_ctx, self.ctx)
        self._js_watch_parser = JsWatchParser(self.ctx)
        self._expr_watch_parser = ExprWatchParser(self.ctx)

    def build_component_ref(self, scope: Scope) -> ast_core.ComponentRef:
        cid = self.ctx.scope_mapper.get_or_create_id(scope)

        options = ast_core.ComponentRefVFor()

        for var in scope._injected:
            if isinstance(var, VForIndex):
                options.index.append(self.ctx.var_mapper.get_or_create_id(var))
                continue

            if isinstance(var, VForItem):
                options.value.append(self.ctx.var_mapper.get_or_create_id(var))
                continue

            if isinstance(var, VForItemKey):
                options.item_key.append(self.ctx.var_mapper.get_or_create_id(var))
                continue

        return ast_core.ComponentRef(id=cid, vfor=options or None)

    def build_component_def(self, scope: Scope) -> ast_core.ComponentDef:
        cid = self.ctx.scope_mapper.get_or_create_id(scope)
        comp = ast_core.ComponentDef(id=cid)

        with self.ctx.component_scope_manager.push(comp):
            comp.variables.extend(
                self._variable_builder.build(variable)
                for variable in scope.variable_order
            )

            comp.renders.extend(
                self._render_builder.build(render) for render in scope._renderables
            )
            comp.web_watch_tasks = self._web_watch_parser.parse(scope)

            comp.injects = self._parse_injects(scope)
            comp.provides = self._parse_provides(scope)
            comp.js_watchs = self._js_watch_parser.parse(scope)
            comp.expr_watchs = self._expr_watch_parser.parse(scope)

        return comp

    def _parse_injects(self, scope: Scope) -> Optional[list[SymbolId]]:
        if not scope._injected:
            return None

        return [self.ctx.var_mapper.get_or_create_id(var) for var in scope._injected]

    def _parse_provides(self, scope: Scope) -> Optional[list[SymbolId]]:
        if not scope._provided:
            return None

        return [self.ctx.var_mapper.get_or_create_id(var) for var in scope._provided]



from typing import TypeVar, cast
from .bindable import VarableBindHelper, BindableMixin
from .app_context import get_current_scope

_TOutput = TypeVar("_TOutput")


class ToValue(BindableMixin):
    def __init__(self, value) -> None:
        self.value = value
        self._bind_helper = VarableBindHelper(
            self,
            define_scope=get_current_scope(),
            lazy_mark_used=[value],
        )

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        pass


def to_value(value: _TOutput) -> _TOutput:
    """
    Converts a reactive state or computed value to a static value.

    Args:
        value (_TOutput): A reactive state or computed value to be converted to its static equivalent.

    Example:
    .. code-block:: python
        org = ui.state("foo")

        # The initial value is "foo", but it will not sync with org.
        other = ui.state(ui.to_value(org))
    """
    return cast(_TOutput, ToValue(value))



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Optional, Union
from typing_extensions import Unpack
from instaui.systems import file_path_system
from instaui.internal.runtime.webview.config import PageBuildStrategy

if TYPE_CHECKING:
    from instaui.internal.runtime.webview.options import WebviewStartArgs


class WebViewLauncher:
    """Example usage:
    .. code-block:: python
        from instaui import ui

        @ui.page("/")
        def index_page():
            ui.content("Hello, world!")

        ui.webview().run()
    """

    def __init__(
        self,
        *,
        assets_path: Union[str, Path] = "webview_assets",
        debug: bool = False,
        auto_create_window: Union[bool, str] = "/",
        on_page_mounted: Optional[Callable] = None,
        clean_assets_on_start: bool = True,
        page_build_strategy: PageBuildStrategy = PageBuildStrategy.LAZY,
        prebuild_routes: Optional[list[str]] = None,
    ):
        """Create a new webview wrapper.

        Args:
            assets_path (Path): Path to store assets.
            debug (bool, optional): Whether to run in debug mode. Defaults to False.
            auto_create_window (Union[bool, str], optional): Whether to create a window automatically. If a string is provided, it will be used as the initial page URL. Defaults to "/".
            on_page_mounted (Optional[Callable[[webview.Window], None]], optional): A callback function to be called when the page is mounted. Defaults to None.
            clean_assets_on_start (bool, optional): Whether to clean the assets folder on start. Defaults to True.
            page_build_strategy (PageBuildStrategy, optional): Controls the default page generation behavior for routes not explicitly prebuilt at launch time. Defaults to PageBuildStrategy.LAZY.
            prebuild_routes (Optional[list[str]], optional): An optional list of routes to be generated eagerly at launch, regardless of the default page build strategy. Defaults to None.
        """
        from instaui.internal.runtime.webview.webview_runtime import WebViewRuntime
        from instaui.internal.runtime.webview.config import WebViewUserLaunchConfig

        assets_path = file_path_system.get_caller_path().parent.joinpath(assets_path)

        self.launch_config = WebViewUserLaunchConfig(
            assets_path=assets_path,
            debug=debug,
            auto_create_window=auto_create_window,
            on_page_mounted=on_page_mounted,
            clean_assets_on_start=clean_assets_on_start,
            page_build_strategy=page_build_strategy,
            prebuild_routes=prebuild_routes,
        )

        self._runtime = WebViewRuntime(self.launch_config)

    def create_window(
        self,
        page_url: str = "/",
    ):
        """Create a new window. Returns the window object of pywebview.

        Args:
            page_url (str, optional): Page URL to load. Defaults to "/".

        """
        return self._runtime.create_window(page_url)

    def run(self, **webview_start_args: Unpack[WebviewStartArgs]):
        """Run the webview.

        Args:
            :param func: Function to invoke upon starting the GUI loop.
            :param args: Function arguments. Can be either a single value or a tuple of
                values.
            :param localization: A dictionary with localized strings. Default strings
                and their keys are defined in localization.py.
            :param gui: Force a specific GUI. Allowed values are ``cef``, ``qt``,
                ``gtk``, ``mshtml`` or ``edgechromium`` depending on a platform.
            :param http_server: Enable built-in HTTP server. If enabled, local files
                will be served using a local HTTP server on a random port. For each
                window, a separate HTTP server is spawned. This option is ignored for
                non-local URLs.
            :param user_agent: Change user agent string.
            :param private_mode: Enable private mode. In private mode, cookies and local storage are not preserved.
                Default is True.
            :param storage_path: Custom location for cookies and other website data
            :param menu: List of menus to be included in the app menu
            :param server: Server class. Defaults to BottleServer
            :param server_args: dictionary of arguments to pass through to the server instantiation
            :param ssl: Enable SSL for local HTTP server. Default is False.
            :param icon: Path to the icon file. Supported only on GTK/QT.
        """
        from instaui.internal.runtime.webview.options import WebViewUserRunOptions

        options = WebViewUserRunOptions(webview_start_args=webview_start_args)
        return self._runtime.run(options)

    def _run_with_index(
        self, index_page: Path, webview_start_args: WebviewStartArgs | None = None
    ):
        from instaui.internal.runtime.webview.options import WebViewUserRunOptions

        options = WebViewUserRunOptions(webview_start_args=webview_start_args or {})
        url = str(index_page.absolute())
        return self._runtime._run_with_index(url, options)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from instaui.internal.backend.invocation import BackendInvocation


class BaseBackend(ABC):
    @abstractmethod
    def register_invocation(
        self, ref_id: int, invocation: BackendInvocation
    ) -> Any: ...


class AssetBackendMixin(ABC):
    @abstractmethod
    def register_asset(self, path: Path) -> str: ...



from typing import Optional

from instaui.systems.dataclass_system import dataclass
from .event_modifier import TEventModifier
from .event import EventMixin


@dataclass()
class EventArgs:
    name: str
    event: EventMixin
    extends: Optional[list] = None
    modifier: Optional[list[TEventModifier]] = None



"""
Preset modules and factory functions for quick usage.
"""

from typing import Optional
from instaui.constants.ui import (
    STORAGE_REF_METHOD,
    INSTAUI_JS_MODULE_NAME,
    USE_DARK_REF_METHOD,
    USE_LANGUAGE_METHOD,
)
from instaui.systems.dataclass_system import dataclass


@dataclass()
class ImportItem:
    module_name: str
    member_name: str
    member_alias: Optional[str] = None


class Vue:
    """
    Preset factory for Vue imports.

    Usage:
        Vue.ref()                (module='vue', name='ref')
        Vue.ref(alias="r")       (module='vue', name='ref', alias='r')
        Vue.reactive()           (module='vue', name='reactive')
    """

    module_name = "vue"

    @staticmethod
    def ref(alias: Optional[str] = None):
        return ImportItem(
            module_name=Vue.module_name,
            member_name="ref",
            member_alias=alias,
        )

    @staticmethod
    def to_refs(alias: Optional[str] = None):
        return ImportItem(
            module_name=Vue.module_name,
            member_name="toRefs",
            member_alias=alias,
        )

    @staticmethod
    def on_mounted():
        return ImportItem(
            module_name=Vue.module_name,
            member_name="onMounted",
        )


class Instaui:
    """
    Preset factory for Instaui imports.
    """

    module_name = INSTAUI_JS_MODULE_NAME

    @staticmethod
    def install():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="install",
        )

    @staticmethod
    def inject():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="inject",
        )

    @staticmethod
    def provide():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="provide",
        )

    @staticmethod
    def ref():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="createRef",
            member_alias="cref",
        )

    @staticmethod
    def render_component():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="renderComponent",
            member_alias="rh",
        )

    @staticmethod
    def render_scope():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="renderScope",
            member_alias="rs",
        )

    @staticmethod
    def render_content():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="content",
            member_alias="rc",
        )

    @staticmethod
    def storage_ref():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name=STORAGE_REF_METHOD,
        )

    @staticmethod
    def use_dark_ref():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name=USE_DARK_REF_METHOD,
        )

    @staticmethod
    def use_language_ref():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name=USE_LANGUAGE_METHOD,
        )

    @staticmethod
    def use_page_title_ref():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="usePageTitleRef",
        )

    @staticmethod
    def web_computed_ref():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="createWebComputedRef",
            member_alias="webCp",
        )

    @staticmethod
    def web_event():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="createWebEvent",
            member_alias="webEvt",
        )

    @staticmethod
    def js_event():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="createJsEvent",
            member_alias="jsEvt",
        )

    @staticmethod
    def expr_event():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="createExprEvent",
            member_alias="exprEvt",
        )

    @staticmethod
    def web_watch_task_scheduler():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="webWatchTaskScheduler",
            member_alias="wts",
        )

    @staticmethod
    def vfor():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="genVFor",
            member_alias="vfor",
        )

    @staticmethod
    def match():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="match",
        )

    @staticmethod
    def vif():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="vif",
        )

    @staticmethod
    def track_path():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="trackPath",
            member_alias="tp",
        )

    @staticmethod
    def js_computed():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="createJsComputed",
            member_alias="jsCp",
        )

    @staticmethod
    def expr_computed():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="createExprComputed",
            member_alias="exprCp",
        )

    @staticmethod
    def js_watch():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="jsWatch",
            member_alias="jws",
        )

    @staticmethod
    def expr_watch():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="exprWatch",
            member_alias="ews",
        )

    @staticmethod
    def str_format():
        return ImportItem(
            module_name=Instaui.module_name,
            member_name="strFormat",
            member_alias="sf",
        )

    @staticmethod
    def to_value():
        return ImportItem(module_name=Instaui.module_name, member_name="toValue")



from pathlib import Path
from typing import Optional
from instaui.systems.dataclass_system import dataclass


@dataclass()
class CdnResourceOption:
    import_maps: Optional[dict[str, str]] = None
    css_links: Optional[dict[Path, str]] = None



import html
from typing import Any, Optional


def dict_to_html_attrs(attrs: Optional[dict[str, Any]]) -> str:
    """
    Convert a dict to an HTML attributes string.

    Example:
        {"id": "app", "class": "container", "disabled": True}
        -> 'id="app" class="container" disabled'
    """
    if not attrs:
        return ""
    parts: list[str] = []

    for key, value in attrs.items():
        if value is None:
            continue

        if value is True:
            parts.append(key)
            continue

        if value is False:
            continue

        escaped_value = html.escape(str(value), quote=True)
        parts.append(f'{key}="{escaped_value}"')

    return " ".join(parts)



from instaui.internal.assets.base import AssetsDeclaration


class PluginImportMapProcessor:
    def process(self, assets: AssetsDeclaration) -> None:
        for plugin in assets.plugins:
            for k, v in plugin.externals.items():
                assets.import_maps.setdefault(k, v)



from __future__ import annotations
from instaui.internal.pipeline.ast_build.base import RenderBuilderProtocol
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder

from instaui.internal.ui.match import Match
from instaui.internal.ast import core as ast_core
from instaui.internal.pipeline.ast_build.context import AstBuilderContext


class MatchParser:
    def __init__(
        self, ctx: AstBuilderContext, render_builder: RenderBuilderProtocol
    ) -> None:
        self.ctx = ctx
        self._expr_builder = ExprBuilder(ctx)
        self._render_builder = render_builder

    def parse(self, match: Match) -> ast_core.MatchRender:
        cond = self._expr_builder.build(match._cond, "literal")

        cases = [
            ast_core.MatchCase(
                value=self._expr_builder.build(case._value, "literal"),
                children=[self._render_builder.build(r) for r in case._renderables],
            )
            for case in match._cases
        ]

        default_case = (
            [self._render_builder.build(r) for r in match._default_case._renderables]
            if match._default_case
            else None
        )

        return ast_core.MatchRender(
            condition=cond,
            cases=cases,
            default_case=default_case,
        )



from __future__ import annotations
from typing import Any

from instaui.internal.ui.app_context import get_current_container
from instaui.internal.ui.renderable import Renderable
from instaui.internal.ui.bindable import mark_used


class Content(Renderable):
    def __init__(self, content: Any):
        """Content to be displayed on the page, typically used for pure text content within slots.

        Args:
            content (Any): The textual content to display.

        Examples:
        .. code-block:: python
            with html.div():
                ui.content("Hello, world!")
        """
        mark_used(content)
        self._content = content
        get_current_container().add_child(self)



from __future__ import annotations
from typing import Optional
from instaui.constants.runtime import RuntimeMode
from instaui.systems.dataclass_system import (
    dataclass,
    asdict_no_none,
    metadata_key_resolver,
    replace,
    field,
)


@dataclass()
class AppMeta:
    mode: RuntimeMode
    version: str
    debug: bool
    route: Optional[ServerRoute] = None
    server_info: Optional[ServerInfo] = field(
        default=None, metadata={"key": "serverInfo"}
    )
    app_icons: Optional[AppIcons] = field(default=None, metadata={"key": "appIcons"})

    def to_dict(self):
        return asdict_no_none(self, key_resolver=metadata_key_resolver())

    def replace(
        self,
        *,
        app_icons: Optional[AppIcons] = None,
        route: Optional[ServerRoute] = None,
    ):
        return replace(self, app_icons=app_icons, route=route)


@dataclass()
class ServerInfo:
    watch_url: str
    watch_async_url: str
    event_url: str
    event_async_url: str
    download_url: str
    assets_url: str
    assets_icons_name: str


@dataclass()
class ServerRoute:
    path: str


@dataclass()
class AppIcons:
    names: Optional[list[str]] = None
    sets: Optional[list[str]] = None

    def __bool__(self):
        return bool(self.names or self.sets)



from typing import TYPE_CHECKING
from .version import __version__

from .zero import ZeroLauncher as zero


if TYPE_CHECKING:
    # only for IDE
    from . import facade_ui as ui
    from instaui.internal.ui import html, file_io
else:

    def __getattr__(name: str):
        match name:
            case "ui":
                from . import facade_ui as ui

                return ui
            case "html":
                from instaui.internal.ui import html

                return html
            case "file_io":
                from instaui.internal.ui import file_io

                return file_io
            case _:
                raise AttributeError(name)


__all__ = [
    "__version__",
    "ui",
    "html",
    "file_io",
    "zero",
]



__all__ = [
    "AssetsPipeline",
    "PluginCssProcessor",
    "PluginImportMapProcessor",
    "ZeroModeComponentCssProcessor",
    "ZeroModeComponentImportMapProcessor",
    "ComponentImportMapProcessor",
    "ComponentCssProcessor",
    "ZeroModeComponentExtensionProcessor",
]

from .pipeline import AssetsPipeline
from .processors.plugin_css import PluginCssProcessor
from .processors.plugin_import_map import PluginImportMapProcessor
from .processors.comp_css import ComponentCssProcessor, ZeroModeComponentCssProcessor
from .processors.comp_import_map import (
    ComponentImportMapProcessor,
    ZeroModeComponentImportMapProcessor,
)
from .processors.comp_extensions import ZeroModeComponentExtensionProcessor



from pathlib import Path
from instaui.constants.ui import USER_ASSETS_DIR_NAME, USER_ASSETS_ICONS_DIR_NAME
from instaui.version import __version__ as _INSTA_VERSION
from instaui.resources import (
    APP_CSS_PATH,
    APP_ES_JS_PATH,
    FAVICON_PATH,
    STATIC_DIR,
    VUE_ES_JS_PATH,
)

RUNTIME_API_PREFIX = "/api/_internal"


class ServerEndpoints:
    def __init__(self, prefix: str = ""):
        self.prefix = self._normalize_prefix(prefix)

        self.INSTAUI_STATIC_URL = f"/_instaui_{_INSTA_VERSION}/static"
        self.INSTAUI_STATIC_URL_WITH_PREFIX = self.with_prefix(self.INSTAUI_STATIC_URL)
        self.INSTAUI_USER_ASSETS_URL = f"/_internal/{USER_ASSETS_DIR_NAME}"
        self.INSTAUI_USER_ASSETS_URL_WITH_PREFIX = self.with_prefix(
            self.INSTAUI_USER_ASSETS_URL
        )
        self.INSTAUI_USER_ICONS_URL = (
            f"{self.INSTAUI_USER_ASSETS_URL}/{USER_ASSETS_ICONS_DIR_NAME}"
        )
        self.INSTAUI_USER_ICONS_URL_WITH_PREFIX = self.with_prefix(
            self.INSTAUI_USER_ICONS_URL
        )
        self.STATIC_DIR = STATIC_DIR

        # static files
        self.VUE_JS_HASH_LINK = self._build_static_url(VUE_ES_JS_PATH)
        self.INSTAUI_JS_HASH_LINK = self._build_static_url(APP_ES_JS_PATH)
        self.APP_CSS_LINK = self._build_static_url(APP_CSS_PATH)
        self.FAVICON_LINK = self._build_static_url(FAVICON_PATH)

        self.ASSETS_URL = f"/_instaui_{_INSTA_VERSION}/resource"
        self.ASSETS_URL_WITH_PREFIX = self.with_prefix(self.ASSETS_URL)
        self.FILE_ROUTER_URL = (
            f"{self.ASSETS_URL}/{{hash_part:path}}/{{file_name:path}}"
        )

        # API
        self.WATCH_URL = f"{RUNTIME_API_PREFIX}/watch/sync"
        self.ASYNC_WATCH_URL = f"{RUNTIME_API_PREFIX}/watch/async"
        self.EVENT_URL = f"{RUNTIME_API_PREFIX}/event/sync"
        self.ASYNC_EVENT_URL = f"{RUNTIME_API_PREFIX}/event/async"
        self.DOWNLOAD_URL = f"{RUNTIME_API_PREFIX}/download"
        self.UPLOAD_URL = f"{RUNTIME_API_PREFIX}/upload_file"

        self.WATCH_URL_WITH_PREFIX = self.with_prefix(self.WATCH_URL)
        self.ASYNC_WATCH_URL_WITH_PREFIX = self.with_prefix(self.ASYNC_WATCH_URL)
        self.EVENT_URL_WITH_PREFIX = self.with_prefix(self.EVENT_URL)
        self.ASYNC_EVENT_URL_WITH_PREFIX = self.with_prefix(self.ASYNC_EVENT_URL)
        self.DOWNLOAD_URL_WITH_PREFIX = self.with_prefix(self.DOWNLOAD_URL)
        self.UPLOAD_URL_WITH_PREFIX = self.with_prefix(self.UPLOAD_URL)

    @staticmethod
    def _normalize_prefix(prefix: str) -> str:
        if not prefix or prefix == "/":
            return ""
        return "/" + prefix.strip("/")

    def with_prefix(self, path: str) -> str:
        if not path.startswith("/"):
            path = "/" + path
        return f"{self.prefix}{path}"

    def _build_static_url(self, target_path: Path) -> str:
        return f"{self.INSTAUI_STATIC_URL}/{target_path.relative_to(self.STATIC_DIR)}"



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from instaui.internal.backend.base import BaseBackend
from instaui.internal.backend.invocation import BackendInvocation


class BackendBindingRegistryBase(ABC):
    @abstractmethod
    def register(self, invocation: BackendInvocation) -> int:
        pass

    @abstractmethod
    def register_asset(self, path: Path) -> int:
        pass

    @abstractmethod
    def get_value(self, ref_id: int) -> Any:
        pass


class BackendBindingRegistry(BackendBindingRegistryBase):
    def __init__(self, backend: BaseBackend):
        self.backend = backend
        self._counter = 0
        self._asset_refs: dict[int, Path] = {}
        self._value_map: dict[int, Any] = {}

    def register(self, invocation: BackendInvocation) -> int:
        rid = self._next_id()
        self._value_map[rid] = self.backend.register_invocation(rid, invocation)
        return rid

    def register_asset(self, path: Path) -> int:
        rid = self._next_id()
        self._asset_refs[rid] = path
        return rid

    def get_value(self, ref_id: int) -> Any:
        return self._value_map[ref_id]

    def _next_id(self) -> int:
        self._counter += 1
        return self._counter


class UnboundBackendBindingRegistry(BackendBindingRegistryBase):
    def register(self, invocation: BackendInvocation) -> int:
        raise NotImplementedError

    def register_asset(self, path: Path) -> int:
        raise NotImplementedError

    def get_value(self, ref_id: int) -> Any:
        raise NotImplementedError



from typing import Any
from instaui.internal.ui.bindable import BindableMixin


class PropInjectableRef(BindableMixin):
    def __init__(self, value: Any):
        self.value = value

    @property
    def _used(self) -> bool:
        return True

    def _mark_used(self) -> None:
        pass

    def _mark_provided(self) -> None:
        pass


def refs(value: Any):
    """
    Convert reactive values contained in an arbitrary structure into
    injectable references, while keeping their positions as placeholders
    in the original structure.

    This function is primarily used at the framework and component-definition
    level to describe *prop-scoped injectable references*. The returned object
    is expected to be bindable and can later be resolved or injected by the
    runtime when the component is instantiated.

    Args:
        value (Any):
            An arbitrary Python structure (e.g. dict, list, tuple, or nested
            combinations thereof) that may contain reactive or bindable values.

    Examples:
    .. code-block:: python
        class MyComponent(custom.element):
            def __init__(self, props):
                super().__init__()
                refs = [
                    {'path': ['a', 'b'], 'bindable': custom.convert_reference(data['a']['b'])}
                ]

                self.props({"refs": custom.refs(refs)})


    ## js component example
    .. code-block:: js
        import { h } from 'vue'
        import { useBindingGetter } from "instaui";

        export default {
            props: ['opts', 'deps'],
            setup(props) {
                const { getValue, getRef } = useBindingGetter();

                // get the ref object
                const refObj = getRef(props.refs[0].bindable)

                return () => h('p', {}, getValue(props.deps[0].bindable))
            }
        }
    """
    return PropInjectableRef(value)



from instaui.internal import import_presets
from instaui.internal.ast import expression
from instaui.internal.ast.core import ComponentRef
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.expr_codegen import ExpressionCodegen


class ComponentRefRenderer:
    def __init__(self, ctx: FileCodegenContext) -> None:
        self.ctx = ctx
        self._expr_builder = ExpressionCodegen(ctx)

    def emit(self, component_ref: ComponentRef) -> str:
        method = self.ctx.imports.use_from_preset(import_presets.Instaui.render_scope())

        vfor_options = {}
        if component_ref.vfor:
            if component_ref.vfor.index:
                vfor_options["index"] = expression.ListExpr.from_values(
                    *[
                        expression.ListExpr.from_values(
                            *[
                                expression.Literal(self.ctx.names.resolve(target)),
                                expression.IdentifierRef(target),
                            ]
                        )
                        for target in component_ref.vfor.index
                    ]
                )

            if component_ref.vfor.value:
                vfor_options["value"] = expression.ListExpr.from_values(
                    *[
                        expression.ListExpr.from_values(
                            *[
                                expression.Literal(self.ctx.names.resolve(target)),
                                expression.IdentifierRef(target),
                            ]
                        )
                        for target in component_ref.vfor.value
                    ]
                )

            if component_ref.vfor.item_key:
                vfor_options["itemKey"] = expression.ListExpr.from_values(
                    *[
                        expression.ListExpr.from_values(
                            *[
                                expression.Literal(self.ctx.names.resolve(target)),
                                expression.IdentifierRef(target),
                            ]
                        )
                        for target in component_ref.vfor.item_key
                    ]
                )

        args = [self.ctx.names.resolve(component_ref.id)]
        if vfor_options:
            opt = expression.ObjectExpr.from_dict(
                {"vfor": expression.ObjectExpr.from_dict(vfor_options)}
            )
            args.append(self._expr_builder.emit(opt))

        return f"{method}({', '.join(args)})"



import contextvars
from contextlib import contextmanager
from typing import Any, Callable
import logging

logger = logging.getLogger(__name__)

_IN_PAGE = contextvars.ContextVar("_IN_PAGE", default=False)

_PAGE_EXIT_CALLBACKS: contextvars.ContextVar[list[Callable[[], None]] | None] = (
    contextvars.ContextVar("_PAGE_EXIT_CALLBACKS", default=None)
)


def in_page_context() -> bool:
    """Returns whether currently in the execution context of page()"""
    return _IN_PAGE.get()


def on_page_exit(callback: Callable[[], Any]) -> None:
    """
    Register a callback to be invoked when exiting the current page context.

    Must be called inside a page context.
    """
    if not in_page_context():
        raise RuntimeError(
            "on_page_exit() must be called inside a page() execution context"
        )

    callbacks = _PAGE_EXIT_CALLBACKS.get()
    if callbacks is None:
        raise RuntimeError("Page exit callback storage is not initialized")

    callbacks.append(callback)


@contextmanager
def enter_page_context():
    """
    Enters the page context.
    The page() decorator should call this context manager before invoking the user function.
    """
    token = _IN_PAGE.set(True)
    cb_token = _PAGE_EXIT_CALLBACKS.set([])
    try:
        yield
    finally:
        callbacks = _PAGE_EXIT_CALLBACKS.get() or []
        for cb in callbacks:
            try:
                cb()
            except Exception:
                logger.exception("Error in page exit callback")

        _PAGE_EXIT_CALLBACKS.reset(cb_token)
        _IN_PAGE.reset(token)



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from instaui.internal.assets.component_extension_registry import (
    ComponentExtensionRegistry,
)
from instaui.systems.dataclass_system import dataclass, field
from .css_assets import CssAssetCollection
from .style_assets import StyleTag
from .component_dep import ComponentDependencyRegistry

if TYPE_CHECKING:
    from .script_assets import ScriptTag
    from .plugin import PluginDependencyInfo


@dataclass
class AssetsDeclaration:
    js_links: set[Union[str, Path]] = field(default_factory=set)
    css_links: CssAssetCollection = field(default_factory=CssAssetCollection)
    style_tags: list[StyleTag] = field(default_factory=list)
    script_tags: list[ScriptTag] = field(default_factory=list)
    import_maps: dict[str, Union[str, Path]] = field(default_factory=dict)
    favicon: Optional[Union[str, Path]] = field(default=None)
    plugins: set[PluginDependencyInfo] = field(default_factory=set)
    component_dependencies: ComponentDependencyRegistry = field(
        default_factory=ComponentDependencyRegistry
    )
    component_extensions: ComponentExtensionRegistry = field(
        default_factory=ComponentExtensionRegistry
    )

    def clone(self) -> AssetsDeclaration:
        """
        Use only when necessary. Typically, the global and local can be merged via the pipeline.
        """
        import copy

        return copy.deepcopy(self)



from typing import Any, Literal, Sequence
from instaui.internal.backend.invocation import BackendInvocation, BackendInvocationKind
from instaui.internal.ui._expr.codec_expr import DataCodecExpr
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui._expr.maybe_ref_expr import UIMaybeRefExpr
from instaui.internal.ui.file_io._download import DownloadFileBindingOutput
from instaui.internal.ui.js_output import JsOutput
from instaui.internal.ui.path_var import PathTrackerBindable, PathTrackerPipelines
from instaui.internal.ui.reference import VariableReferenceName
from instaui.internal.ui.to_value import ToValue
from instaui.internal.ui.upload import UploadEndpoint
from instaui.internal.ui.variable import Variable
from instaui.internal.ui.str_format import StrFormat
from instaui.internal.ui.event_context import EventContext
from instaui.internal.ui.props_injectable import PropInjectableRef
from instaui.internal.ast import expression
from instaui.internal.pipeline.ast_build.context import AstBuilderContext
from instaui.internal.codec.codec_json import json_codec


TBuildMode = Literal["expr", "literal", "shallow"]


class ExprBuilder:
    def __init__(self, ctx: AstBuilderContext) -> None:
        self.ctx = ctx

    def build(self, value: Any, mode: TBuildMode = "expr") -> expression.Expression:
        method = getattr(self, f"build_{type(value).__name__}", None)
        if method:
            return method(value, mode)

        if isinstance(value, PathTrackerBindable):
            return self.build_PathTrackerBindable(value, mode)
        if isinstance(value, Variable):
            return self.build_Variable(value, mode)

        if isinstance(value, JsOutput):
            return self.build_JsOutput(value, mode)

        if isinstance(value, (str, int, float, bool)) or value is None:
            return expression.Literal(value=value)

        if isinstance(value, (list, tuple)):
            return self._build_sequence(value, mode)

        if isinstance(value, dict):
            return self._build_dict(value, mode)

        if isinstance(value, expression.Expression):
            return value

        raise TypeError(f"Unsupported expression value: {value!r}")

    def _build_sequence(
        self, value: Sequence, mode: TBuildMode
    ) -> expression.Expression:
        if mode == "expr":
            return expression.ListExpr([self.build(v, mode) for v in value])

        if mode == "literal":
            return expression.ListLiteral(value=list(value))

        if mode == "shallow":
            return self._build_shallow_literal(value)

        raise AssertionError(mode)

    def _build_dict(self, value: dict, mode: TBuildMode) -> expression.Expression:
        if mode == "expr":
            return expression.ObjectExpr(
                props=[
                    expression.ObjectProperty(k, self.build(v, mode))
                    for k, v in value.items()
                ]
            )

        if mode == "literal":
            return expression.ObjectLiteral(props=value)

        if mode == "shallow":
            return self._build_shallow_literal(value)

        raise AssertionError(mode)

    def _build_shallow_literal(self, value: Any) -> expression.Expression:
        if isinstance(value, expression.Expression):
            return value

        if isinstance(value, list):
            return expression.ListLiteral(value=list(value))

        if isinstance(value, dict):
            return expression.ObjectLiteral(props=dict(value))

        if isinstance(value, (str, int, float, bool)) or value is None:
            return expression.Literal(value=value)

        raise TypeError(f"Unsupported value in shallow literal: {value!r}")

    def object_shallow(self, kwargs: dict) -> expression.ObjectExpr:
        return expression.ObjectExpr(
            props=[
                expression.ObjectProperty(k, self.build(v, "shallow"))
                for k, v in kwargs.items()
                if v is not None
            ]
        )

    def object_expr(
        self,
        kwargs: dict,
        *,
        mode: TBuildMode = "expr",
    ) -> expression.ObjectExpr:
        return expression.ObjectExpr(
            props=[
                expression.ObjectProperty(k, self.build(v, mode))
                for k, v in kwargs.items()
                if v is not None
            ]
        )

    def list_expr(
        self, *, mode: TBuildMode = "expr", values: Sequence[Any]
    ) -> expression.ListExpr:
        return expression.ListExpr([self.build(v, mode) for v in values])

    def list_shallow(self, values: Sequence[Any]) -> expression.ListExpr:
        return expression.ListExpr([self.build(v, "shallow") for v in values])

    def build_Variable(
        self, value: Variable, mode: TBuildMode = "expr"
    ) -> expression.IdentifierRef:
        real_value = value._source if isinstance(value, PathTrackerBindable) else value
        sid = self.ctx.var_mapper.get_id(real_value)
        assert sid is not None, f"Undefined variable: {real_value!r}"
        return expression.IdentifierRef(id=sid)

    def build_PathTrackerBindable(
        self, value: PathTrackerBindable, mode: TBuildMode = "expr"
    ):
        target = self.build(value._source)
        args = self.build(PathTrackerPipelines.to_expr_literal(value))
        return expression.PathTrackerBindableCall(target, args)

    def build_StrFormat(self, value: StrFormat, mode: TBuildMode):
        code = self.build(value.template)
        args = expression.ObjectExpr.from_kwargs(
            code=code,
            args=self.list_shallow(value.args) if value.args else None,
            kws=self.object_shallow(value.kwargs) if value.kwargs else None,
        )

        return expression.StrFormatCall(args)

    def build_EventContext(self, value: EventContext, mode: TBuildMode):
        code = value.path[1:] if value.path[0] == ":" else f"e=> e.{value.path}"
        return expression.RawFunctionExpr(code)

    def build_JsOutput(self, value: JsOutput, mode: TBuildMode):
        return expression.NULL

    def build_DownloadFileBindingOutput(
        self, value: DownloadFileBindingOutput, mode: TBuildMode
    ):
        return expression.NULL

    def build_UploadEndpoint(self, value: UploadEndpoint, mode: TBuildMode):
        ref_id = self.ctx.binding_registry.register(
            BackendInvocation(
                kind=BackendInvocationKind.FILE_UPLOAD,
                fn=value.handler,
                render_ctx=self.ctx.render_ctx,
            )
        )

        return expression.BackendValueRefExpr(ref_id)

    def build_VariableReferenceName(
        self, value: VariableReferenceName, mode: TBuildMode
    ):
        sid = self.ctx.var_mapper.get_id(value.variable)
        assert sid is not None, f"Undefined variable: {value.variable!r}"
        return expression.StringLiteral(expression.IdentifierRef(id=sid))

    def build_PropInjectableRef(self, value: PropInjectableRef, mode: TBuildMode):
        return self.build(value.value, "expr")

    def build_ToValue(self, value: ToValue, mode: TBuildMode):
        return expression.ToValueCall(self.build(value.value))

    def build_UILiteralExpr(self, value: UILiteralExpr, mode: TBuildMode):
        return expression.JsonLiteralExpr(value.value)

    def build_UIMaybeRefExpr(self, value: UIMaybeRefExpr, mode: TBuildMode):
        return self.build(value.value, mode)

    def build_DataCodecExpr(self, expr: DataCodecExpr, mode: TBuildMode):
        return expression.JsonLiteralExpr(json_codec.python_to_json_value(expr.value))



from typing import Literal, Union
from typing_extensions import TypedDict
from instaui.internal.ui.components._responsive_type._common import (
    TMaybeResponsive,
    TLevel_0_9,
    TLevel_neg_9_9,
    TWithAutoValue,
)
from instaui.internal.ui.components._responsive_type._layout import TOverflowEnum


class TLayoutBaseProps(TypedDict, total=False):
    p: TMaybeResponsive[Union[str, TLevel_0_9, TWithAutoValue]]
    px: TMaybeResponsive[Union[str, TLevel_0_9, TWithAutoValue]]
    py: TMaybeResponsive[Union[str, TLevel_0_9, TWithAutoValue]]
    pt: TMaybeResponsive[Union[str, TLevel_0_9, TWithAutoValue]]
    pb: TMaybeResponsive[Union[str, TLevel_0_9, TWithAutoValue]]
    pl: TMaybeResponsive[Union[str, TLevel_0_9, TWithAutoValue]]
    pr: TMaybeResponsive[Union[str, TLevel_0_9, TWithAutoValue]]
    width: TMaybeResponsive[str]
    height: TMaybeResponsive[str]
    min_width: TMaybeResponsive[str]
    min_height: TMaybeResponsive[str]
    max_width: TMaybeResponsive[str]
    max_height: TMaybeResponsive[str]
    position: TMaybeResponsive[
        Literal["static", "relative", "absolute", "fixed", "sticky"]
    ]
    inset: TMaybeResponsive[Union[str, TLevel_neg_9_9]]
    top: TMaybeResponsive[Union[str, TLevel_neg_9_9]]
    right: TMaybeResponsive[Union[str, TLevel_neg_9_9]]
    bottom: TMaybeResponsive[Union[str, TLevel_neg_9_9]]
    left: TMaybeResponsive[Union[str, TLevel_neg_9_9]]
    overflow: TMaybeResponsive[Union[TOverflowEnum, str]]
    overflow_x: TMaybeResponsive[Union[TOverflowEnum, str]]
    overflow_y: TMaybeResponsive[Union[TOverflowEnum, str]]
    flex_basis: TMaybeResponsive[str]
    flex_shrink: TMaybeResponsive[Union[str, Literal["0", "1"]]]
    flex_grow: TMaybeResponsive[Union[str, Literal["0", "1"]]]
    grid_area: TMaybeResponsive[str]
    grid_column: TMaybeResponsive[str]
    grid_column_start: TMaybeResponsive[str]
    grid_column_end: TMaybeResponsive[str]
    grid_row: TMaybeResponsive[str]
    grid_row_start: TMaybeResponsive[str]
    grid_row_end: TMaybeResponsive[str]
    m: TMaybeResponsive[Union[str, TLevel_neg_9_9, TWithAutoValue]]
    mx: TMaybeResponsive[Union[str, TLevel_neg_9_9, TWithAutoValue]]
    my: TMaybeResponsive[Union[str, TLevel_neg_9_9, TWithAutoValue]]
    mt: TMaybeResponsive[Union[str, TLevel_neg_9_9, TWithAutoValue]]
    mr: TMaybeResponsive[Union[str, TLevel_neg_9_9, TWithAutoValue]]
    mb: TMaybeResponsive[Union[str, TLevel_neg_9_9, TWithAutoValue]]
    ml: TMaybeResponsive[Union[str, TLevel_neg_9_9, TWithAutoValue]]



from typing import Mapping, Optional
from instaui.protocol.codec.type_adapter import TypeAdapterProtocol
from instaui.systems.dataclass_system import dataclass


class InvocationSpec:
    """Marker base class for typed invocation specs."""

    pass


@dataclass(frozen=True)
class ComputedSpec(InvocationSpec):
    outputs_binding_count: int
    custom_type_adapter_map: Optional[Mapping[int, TypeAdapterProtocol]] = None


@dataclass(frozen=True)
class WatchSpec(InvocationSpec):
    outputs_binding_count: int
    custom_type_adapter_map: Optional[Mapping[int, TypeAdapterProtocol]] = None


@dataclass(frozen=True)
class EventSpec(InvocationSpec):
    outputs_binding_count: int
    dataset_input_indexs: list[int]
    custom_type_adapter_map: Optional[Mapping[int, TypeAdapterProtocol]] = None



import logging
import sys


def setup_logging(*, debug: bool):
    level = logging.DEBUG if debug else logging.INFO

    logging.basicConfig(
        level=level,
        format="%(asctime)s | %(levelname)-8s | %(name)s | %(message)s",
        datefmt="%H:%M:%S",
        handlers=[
            logging.StreamHandler(sys.stdout),
        ],
    )


def setup_instaui_logging(debug: bool):
    logger = logging.getLogger("instaui")
    logger.setLevel(logging.DEBUG if debug else logging.INFO)

    if not logger.hasHandlers():
        handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(
            "%(asctime)s | %(levelname)-8s | %(name)s | %(message)s"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    logger.propagate = False



from __future__ import annotations
import typing
from instaui.debug.api_boundary import user_api
from instaui.debug.source import get_source_span
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.app_index import new_scope_if_needed
from instaui.internal.ui.bindable import BindableMixin, VarableBindHelper
from instaui.internal.ui.enums import InputBindingType
from instaui.internal.ui.path_var import PathVar
from instaui.internal.ui.protocol import ObservableProtocol
from instaui.internal.ui.variable import Variable


_R_TYPE = typing.TypeVar("_R_TYPE")


class JsComputed(Variable, PathVar, BindableMixin, ObservableProtocol):
    def __init__(
        self,
        *,
        inputs: typing.Optional[typing.Sequence] = None,
        code: str,
        async_init_value: typing.Optional[typing.Any] = None,
        deep_compare_on_input: bool = False,
        tool: typing.Optional[typing.Literal["unwrap_reactive"]] = None,
    ) -> None:
        self._source_span_ = get_source_span()
        new_scope_if_needed()
        define_scope = get_current_scope()
        self.code = code
        self._inputs = UILiteralExpr.try_parse_list(inputs or [])

        self._async_init_value = UILiteralExpr.try_parse(async_init_value)
        self._deep_compare_on_input = deep_compare_on_input
        self._tool = tool

        self._bind_helper = VarableBindHelper(
            self,
            define_scope.register_js_computed,
            define_scope=define_scope,
            lazy_mark_used=self._inputs,
        )

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref


TJsComputed = JsComputed


@typing.overload
def js_computed(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    code: str,
    async_init_value: typing.Optional[typing.Any] = None,
    deep_compare_on_input: bool = False,
) -> typing.Any: ...


@typing.overload
def js_computed(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    code: str,
    async_init_value: typing.Optional[typing.Any] = None,
    deep_compare_on_input: bool = False,
    r_type: typing.Optional[typing.Type[_R_TYPE]] = None,
) -> _R_TYPE: ...


@user_api
def js_computed(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    code: str,
    async_init_value: typing.Optional[typing.Any] = None,
    deep_compare_on_input: bool = False,
    r_type: typing.Optional[typing.Type[_R_TYPE]] = None,
) -> typing.Union[JsComputed, _R_TYPE]:
    """
    A client-side computed property that evaluates JavaScript code to derive reactive values.

    Args:
        inputs (typing.Optional[typing.Sequence], optional): Reactive data sources to monitor.
                                                  Changes to these values trigger re-computation.
        code (str): JavaScript code to execute for computing the value.
        async_init_value (typing.Optional[typing.Any], optional): Initial value to use before first successful async evaluation.

    # Example:
    .. code-block:: python
        a = ui.state(0)

        plus_one = ui.js_computed(inputs=[a], code="a=> a+1")

        html.number(a)
        ui.text(plus_one)
    """

    jc = JsComputed(
        inputs=inputs,
        code=code,
        async_init_value=async_init_value,
        deep_compare_on_input=deep_compare_on_input,
    )

    if r_type is None:
        return jc

    return typing.cast(_R_TYPE, jc)



from __future__ import annotations
from typing import Any, Union
from instaui.internal.ui.element import Element


class Li(Element):
    def __init__(
        self,
        text: Union[Any, None] = None,
    ):
        super().__init__("li")

        if text:
            self.props({"innerText": text})



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING
from instaui.systems import file_system

if TYPE_CHECKING:
    from instaui.internal.backend.fastapi.endpoints import ServerEndpoints


_THashPart = str
_HASH_PART_MAP: dict[_THashPart, Path] = {}
_PATH_URL_MAP: dict[Path, _THashPart] = {}


def get_folder_path(hash_part: str) -> Path:
    return _HASH_PART_MAP[hash_part]


def record_asset(path: Path, endpoints: ServerEndpoints):
    path = Path(path).resolve()
    is_file = path.is_file()

    folder_path = path.parent if is_file else path

    if folder_path not in _HASH_PART_MAP:
        hash_part = file_system.generate_hash_name_from_path(folder_path)
        _HASH_PART_MAP[hash_part] = folder_path
    else:
        hash_part = _PATH_URL_MAP[folder_path]

    folder_url = f"{endpoints.ASSETS_URL_WITH_PREFIX}/{hash_part}/"

    return f"{folder_url}{path.name}" if is_file else folder_url



from __future__ import annotations
from typing import Any, Callable
from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware
from instaui.constants.runtime import RuntimeMode
from instaui.internal.ui.app_index import new_app


class RequestContextMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, debug: bool = False):
        super().__init__(app)
        self.debug = debug

    async def dispatch(self, request: Request, call_next: Callable) -> Any:
        with new_app(RuntimeMode.WEB, debug=self.debug):
            return await call_next(request)


class NoCacheDebugModeMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next: Callable) -> Any:
        response = await call_next(request)

        if request.url.path.endswith((".js", ".css")):
            response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
            response.headers["Pragma"] = "no-cache"
            response.headers["Expires"] = "0"

        return response



from __future__ import annotations

import asyncio
import inspect
import logging
import multiprocessing
import os
from pathlib import Path
import time
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Final, Optional, Set, cast

import uvicorn
from fastapi import APIRouter, FastAPI, Request
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from uvicorn.supervisors import ChangeReload

import __main__
from instaui.internal.backend.base import AssetBackendMixin, BaseBackend
from instaui.internal.backend.common import python_handler_utils
from instaui.internal.backend.fastapi.endpoints import ServerEndpoints
from instaui.internal.backend.fastapi.page_cache import PageCache
from instaui.internal.backend.invocation import BackendInvocation, BackendInvocationKind
from instaui.internal.router.model import PageInfo
from instaui.internal.router.registry import route_registry
from instaui.internal.runtime.logging import setup_instaui_logging
from instaui.protocol.invocation.spec import ComputedSpec, EventSpec

from . import (
    asset_mapper,
    debug_mode_router,
    dependency_router,
    event_router,
    file_download_router,
    file_upload_router,
    watch_router,
)
from ._uvicorn import UvicornServer
from .middlewares import NoCacheDebugModeMiddleware, RequestContextMiddleware

if TYPE_CHECKING:
    from instaui.internal.runtime.web.config import WebUserLaunchConfig
    from instaui.internal.runtime.web.options import (
        WebUserRunOptions,
        WebUserRunWithOptions,
    )
    from instaui.internal.runtime.web.runtime_services import WebRuntimeServices

APP_IMPORT_STRING: Final[str] = (
    "instaui.internal.backend.fastapi.web_backend:FastapiBackend.app"
)


class FastapiBackend(BaseBackend, AssetBackendMixin):
    app: Optional[FastAPI] = None

    def __init__(
        self,
        *,
        config: WebUserLaunchConfig,
        user_assets_dir: Optional[Path] = None,
        runtimie_services: WebRuntimeServices,
    ):
        self._config = config
        self._debug_mode = config.debug
        self._use_gzip = config.use_gzip
        self._user_assets_dir = user_assets_dir
        self._use_fns: list[Callable[[FastAPI], None]] = []
        self._router = APIRouter(tags=["InstaUI"], prefix=config.prefix)
        self.endpoints = ServerEndpoints(config.prefix)
        self._runtimie_services = runtimie_services

        dependency_router.create_router(self._router, self.endpoints)
        event_router.create_router(self._router, self._debug_mode, self.endpoints)
        watch_router.create_router(self._router, self._debug_mode, self.endpoints)
        file_upload_router.create_router(self._router, self.endpoints)
        file_download_router.create_router(self._router, self.endpoints)
        debug_mode_router.create_router(self._router, self._debug_mode)

        for path, info in route_registry.routes.items():
            self.register_page(path, info)

        self._registered_static_routes: Set[str] = set()
        self._page_cache = PageCache()
        self._logger = logging.getLogger(__name__)

    @property
    def router(self) -> APIRouter:
        return self.app.router if self.app else self._router

    def register_dynamic_page(self, path: str, fn: Callable, *, cache: bool = False):
        """
        For dynamic page registration after service startup"""
        info = PageInfo(fn, cache)
        route_registry.add_route(path, info)
        self.register_page(path, info)

    def register_page(self, path: str, info: PageInfo):
        """
        For page registration before service startup. After service startup, use register_dynamic_page for registration
        """
        is_async = inspect.iscoroutinefunction(info.fn)

        self._remove_route(path)

        if is_async:

            @self.router.get(path)
            async def _(request: Request):
                html = await self._page_cache.get_or_render_async(
                    key=path,
                    cache=(not self._debug_mode) and info.cache,
                    render=lambda: self._runtimie_services.render_fn_async(path),
                )

                return HTMLResponse(html)

        else:

            @self.router.get(path)
            def _(request: Request):
                html = self._page_cache.get_or_render_sync(
                    key=path,
                    cache=(not self._debug_mode) and info.cache,
                    render=lambda: self._runtimie_services.render_fn(path),
                )

                return HTMLResponse(html)

    def _remove_route(self, path: str) -> None:
        self.router.routes[:] = [
            r for r in self.router.routes if getattr(r, "path", None) != path
        ]

    def try_close_server(self):
        assert UvicornServer.get_instance() is not None, (
            "Uvicorn server does not exist. Cannot attempt to shut down service in debug mode or reload mode"
        )
        UvicornServer.get_instance().should_exit = True

    def use(self, fn: Callable[[FastAPI], None]):
        self._use_fns.append(fn)
        return self

    def run(self, optons: WebUserRunOptions):
        setup_instaui_logging(debug=self._debug_mode)
        reload = self._debug_mode or optons.reload

        app = FastAPI()
        FastapiBackend.app = app
        self._setup_fastapi_app(app)

        if (not self._debug_mode) and self._config.prerender:
            self.prerender_all_pages()

        if multiprocessing.current_process().name != "MainProcess":
            return

        if reload and not hasattr(__main__, "__file__"):
            reload = False

        config = uvicorn.Config(
            APP_IMPORT_STRING if reload else app,
            host=optons.host,
            port=optons.port,
            reload=reload,
            log_level=optons.log_level,
            workers=optons.workers,
            uds=optons.uds,
            reload_includes=_split_args(optons.reload_includes) if reload else None,
            reload_excludes=_split_args(optons.reload_excludes) if reload else None,
            reload_dirs=_split_args(optons.reload_dirs) if reload else None,
            **optons.kwargs,
        )

        UvicornServer.create_singleton(config, [debug_mode_router.when_server_reload])

        if config.should_reload:
            ChangeReload(
                config, target=UvicornServer.get_instance().run, sockets=[]
            ).run()
        else:
            UvicornServer.get_instance().run()

        if config.uds:
            os.remove(config.uds)  # pragma: py-win32

    def run_with(self, options: WebUserRunWithOptions):
        """
        Mounts the InstaUI interface onto a FastAPI application with configurable routing.
        """

        assert isinstance(options.app, FastAPI), "app must be a FastAPI instance"

        if options.tags is not None:
            self._router.tags = options.tags
        if options.dependencies:
            self._router.dependencies = options.dependencies
        if options.responses:
            self._router.responses = options.responses

        self._setup_fastapi_app(options.app, include_router=False)

        if (not self._debug_mode) and self._config.prerender:
            self.prerender_all_pages()

        options.app.include_router(self._router)
        return self._router

    def _setup_fastapi_app(self, app: FastAPI, *, include_router=True):
        app.add_middleware(RequestContextMiddleware, debug=self._debug_mode)
        if self._debug_mode:
            app.add_middleware(NoCacheDebugModeMiddleware)

        if self._use_gzip:
            app.add_middleware(
                GZipMiddleware,
                minimum_size=self._use_gzip if isinstance(self._use_gzip, int) else 500,
            )

        for fn in self._use_fns:
            fn(app)

        app.mount(
            self.endpoints.INSTAUI_STATIC_URL,
            StaticFiles(directory=self.endpoints.STATIC_DIR),
            name=self.endpoints.INSTAUI_STATIC_URL,
        )

        if include_router:
            app.include_router(self._router)

        if self._user_assets_dir:
            app.mount(
                self.endpoints.INSTAUI_USER_ASSETS_URL_WITH_PREFIX,
                StaticFiles(directory=self._user_assets_dir),
            )
            self._logger.info(
                f"User assets directory found: {self._user_assets_dir} : mount to {self.endpoints.INSTAUI_USER_ASSETS_URL}"
            )

    def prerender_all_pages(self):
        """
        Render all cacheable pages before server start.
        """
        routes = [
            (path, info) for path, info in route_registry.routes.items() if info.cache
        ]

        if not routes:
            self._logger.info("prerender skipped: no cacheable pages found")
            return

        self._logger.info("prerender start: %d pages", len(routes))

        async_tasks: list[tuple[str, Awaitable[None]]] = []

        for path, info in routes:
            is_async = inspect.iscoroutinefunction(info.fn)

            if is_async:
                async_tasks.append((path, self._prerender_async_page(path)))
            else:
                self._prerender_sync_page(path)

        if async_tasks:
            self._run_async_prerender(async_tasks)

        self._logger.info("prerender finished: %d pages", len(routes))

    def _prerender_sync_page(self, path: str):
        start = time.perf_counter()
        try:
            html = self._runtimie_services.render_fn(path)
            self._page_cache.warmup(path, html)

            elapsed = (time.perf_counter() - start) * 1000
            self._logger.debug(
                "prerender sync page success: path=%s elapsed=%.2fms",
                path,
                elapsed,
            )

        except Exception:
            self._logger.exception(
                "prerender sync page failed: path=%s",
                path,
            )

    async def _prerender_async_page(self, path: str):
        start = time.perf_counter()
        try:
            html = await self._runtimie_services.render_fn_async(path)
            self._page_cache.warmup(path, html)
            elapsed = (time.perf_counter() - start) * 1000
            self._logger.debug(
                "prerender async page success: path=%s elapsed=%.2fms",
                path,
                elapsed,
            )

        except Exception:
            self._logger.exception(
                "prerender async page failed: path=%s",
                path,
            )

    def _run_async_prerender(self, tasks: list[tuple[str, Awaitable[None]]]):
        async def runner():
            await asyncio.gather(*(task for _, task in tasks))

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        if loop and loop.is_running():
            loop.create_task(runner())
            self._logger.debug("prerender async tasks scheduled in running loop")
        else:
            asyncio.run(runner())

    def register_invocation(self, ref_id: int, invocation: BackendInvocation) -> Any:
        route = invocation.render_ctx.route
        assert route is not None, "route must be set in render_ctx for invocation"
        key = python_handler_utils.create_handler_key(
            route, invocation.fn, self._debug_mode
        )

        if (
            invocation.kind == BackendInvocationKind.COMPUTED
            or invocation.kind == BackendInvocationKind.WATCH
        ):
            spec = cast(ComputedSpec, invocation.spec)
            watch_router.register_watch_handler(
                key,
                invocation.fn,
                spec.outputs_binding_count,
                spec.custom_type_adapter_map,
            )

            return key

        if invocation.kind == BackendInvocationKind.EVENT:
            spec = cast(EventSpec, invocation.spec)
            event_router.register_event_handler(
                key,
                invocation.fn,
                spec.outputs_binding_count,
                spec.dataset_input_indexs,
                spec.custom_type_adapter_map,
            )

            return key

        if invocation.kind == BackendInvocationKind.FILE_UPLOAD:
            file_upload_router.register_upload_file_handler(
                key, invocation.fn, self.endpoints.UPLOAD_URL_WITH_PREFIX
            )

            return f"{self.endpoints.UPLOAD_URL_WITH_PREFIX}?hkey={key}"

    def register_asset(self, path: Path) -> str:
        return asset_mapper.record_asset(path, self.endpoints)


def _split_args(args: str):
    return [a.strip() for a in args.split(",")]



from instaui.internal import import_presets
from instaui.internal.ast.core import (
    VFor,
)
from instaui.internal.ast import expression
from instaui.internal.codegen.base import RenderEmitter
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.expr_codegen import ExpressionCodegen


class VForRenderer:
    def __init__(self, ctx: FileCodegenContext, render_emitter: RenderEmitter) -> None:
        self.ctx = ctx
        self.expr = ExpressionCodegen(ctx)
        self.render_emitter = render_emitter

    def emit(self, vfor: VFor) -> str:
        method = self.ctx.imports.use_from_preset(import_presets.Instaui.vfor())

        array = expression.ObjectExpr.from_kwargs(
            type=vfor.array.type,
            value=vfor.array.value,
        )

        obj = expression.ObjectExpr.from_kwargs(
            fkey=vfor.fkey, array=array, tsGroup=vfor.transition_group
        )

        children_render = self.emit_children(vfor)

        return f"{method}({self.expr.emit(obj)},{self.expr.emit(children_render)})"

    def emit_children(self, vfor: VFor) -> expression.ArrowFunctionExpr:
        if not vfor.children:
            return expression.ArrowFunctionExpr()

        body = [
            expression.RawFunctionExpr(self.render_emitter.emit_render(child))
            for child in vfor.children
        ]

        params = expression.ObjectExpr.from_kwargs(
            i=expression.IdentifierRef(vfor.used_index) if vfor.used_index else None,
            v=expression.IdentifierRef(vfor.used_item) if vfor.used_item else None,
            k=expression.IdentifierRef(vfor.used_key) if vfor.used_key else None,
        )

        return expression.ArrowFunctionExpr(
            params=params, body=expression.ListExpr.from_values(*body)
        )



from __future__ import annotations
from typing import Any, Iterable, Optional, Union
from instaui.internal.ast.property_key import PropertyKey
from instaui.systems.dataclass_system import dataclass, field
from instaui.internal.ast.symbol import SymbolId


class Expression:
    pass


# Member access
@dataclass()
class MemberExpr(Expression):
    """
    a.x
    v.value
    """

    obj: Expression
    member: str


# Function call
@dataclass()
class CallExpr(Expression):
    """
    ref()
    ref(0)
    ref(0, 1)
    createApp(cp, {a:1})
    """

    callee: Expression
    args: tuple[Expression, ...] = field(default_factory=tuple)

    @classmethod
    def of(cls, callee: Expression, *args: Expression) -> CallExpr:
        return cls(callee=callee, args=args)


@dataclass()
class ObjectExpr(Expression):
    props: list[ObjectProperty]

    @classmethod
    def from_kwargs(cls, **kwargs: Optional[Expression]) -> ObjectExpr:
        return cls(
            props=[ObjectProperty(k, v) for k, v in kwargs.items() if v is not None]
        )

    @classmethod
    def from_dict(cls, kwargs: dict[Union[PropertyKey, str], Expression]) -> ObjectExpr:
        return cls(
            props=[ObjectProperty(k, v) for k, v in kwargs.items() if v is not None]
        )

    @classmethod
    def from_pairs(
        cls, pairs: Iterable[tuple[Union[PropertyKey, str], Expression | None]]
    ) -> ObjectExpr:
        return cls(props=[ObjectProperty(k, v) for k, v in pairs if v is not None])

    def __bool__(self) -> bool:
        return bool(self.props)


@dataclass()
class ListExpr(Expression):
    value: list[Expression]

    @classmethod
    def from_values(cls, *values: Expression) -> ListExpr:
        return cls(value=list(values))

    def __bool__(self) -> bool:
        return bool(self.value)


@dataclass()
class ObjectProperty:
    key: Union[PropertyKey, str]
    value: Optional[Expression] = None
    shorthand: bool = False


# Arrow function (for setup / render)
@dataclass()
class ArrowFunctionExpr(Expression):
    body: Optional[Expression] = None
    params: Optional[Expression] = None


@dataclass()
class RawFunctionExpr(Expression):
    """
    Directly embed a string of JavaScript anonymous function.
    For example:
        "function(x) { return x + 1; }"
        "() => ({ count: 0 })"
    Note: The caller must ensure the string is a valid JS function expression.
    """

    js_code: str


@dataclass()
class IdentifierRef(Expression):
    id: SymbolId


# Literal
@dataclass()
class Literal(Expression):
    value: Any


@dataclass()
class StringLiteral(Expression):
    value: Any


@dataclass(frozen=True)
class NullExpr(Expression):
    pass


@dataclass(frozen=True)
class UndefinedExpr(Expression):
    pass


NULL = NullExpr()
UNDEFINED = UndefinedExpr()


@dataclass()
class ObjectLiteral(Expression):
    props: dict[str, Any]

    @classmethod
    def from_kwargs(cls, **kwargs: Any) -> ObjectLiteral:
        return cls(props={k: v for k, v in kwargs.items() if v is not None})

    def __bool__(self) -> bool:
        return bool(self.props)


@dataclass()
class ListLiteral(Expression):
    value: list

    def __bool__(self) -> bool:
        return bool(self.value)


@dataclass(frozen=True)
class JsonLiteralExpr(Expression):
    value: Any

    def __bool__(self) -> bool:
        return bool(self.value)


@dataclass()
class PathTrackerBindableCall(Expression):
    target: Expression
    args: Expression


@dataclass()
class StrFormatCall(Expression):
    args: Expression


@dataclass()
class BackendValueRefExpr(Expression):
    ref_id: int


@dataclass()
class ToValueCall(Expression):
    value: Expression



from abc import ABC, abstractmethod


class Variable(ABC):
    @property
    @abstractmethod
    def _define_scope_id(self) -> int: ...

    def _is_providable(self) -> bool:
        """Indicates whether this variable can be provided. Default: True."""
        return True



from instaui import ui, html
from instaui.internal.ui.protocol import ObservableProtocol


def list_all_bindables(locals_dict: dict):
    """List all bindables in the locals() dictionary.


    Example usage:

    ```python
    list_all_bindables(locals())
    ```

    """

    with html.div().style(
        "display: grid; grid-template-columns: auto 1fr; border: 1px solid black; padding: 10px;"
    ):
        ui.text("variable name")
        ui.text("bindable value").style("justify-self: center;")

        html.div().style(
            "grid-column: 1 / span 2;height: 1px;border-bottom: 1px solid black;"
        )

        for key, value in locals_dict.items():
            if isinstance(value, ObservableProtocol):
                cp_value = ui.js_computed(
                    inputs=[value],
                    code=r"""(obj)=>{
 
    if (typeof obj === 'object') {
        if (obj === null) {
            return 'null';
        } else {
            return JSON.stringify(obj);
        }
    } else {
        return String(obj);
    }
}""",
                )

                html.paragraph(f"{key}:").style("justify-self: end;")
                html.paragraph(cp_value).style("justify-self: center;")



from instaui.internal.ui.enums import OutputSetType
from instaui.internal.ui.protocol import CanOutputProtocol


class JsOutput(CanOutputProtocol):
    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.JsCode



from typing import cast
from contextvars import ContextVar
from ._app import App, DefaultApp


app_var: ContextVar[App] = ContextVar("app_var", default=cast(App, App._default_app))


def get_app() -> App:
    """Get the current App instance from context."""
    app = app_var.get()
    assert app is not None, "No App instance in context"
    return app


def get_current_scope():
    return get_app().get_current_scope()


def get_current_container():
    return get_app().get_current_container()


def get_default_app():
    return cast(DefaultApp, App._default_app)



from instaui.internal import import_presets
from instaui.internal.ast.core import (
    VIf,
)
from instaui.internal.ast import expression
from instaui.internal.codegen.base import RenderEmitter
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.expr_codegen import ExpressionCodegen


class VIfRenderer:
    def __init__(self, ctx: FileCodegenContext, render_emitter: RenderEmitter) -> None:
        self.ctx = ctx
        self.expr = ExpressionCodegen(ctx)
        self.render_emitter = render_emitter

    def emit(self, vif: VIf) -> str:
        method = self.ctx.imports.use_from_preset(import_presets.Instaui.vif())

        children_render = self.emit_children(vif)
        opts = expression.ObjectExpr.from_kwargs(on=vif.condition)
        return f"{method}({self.expr.emit(opts)},{self.expr.emit(children_render)})"

    def emit_children(self, vif: VIf) -> expression.ArrowFunctionExpr:
        if not vif.children:
            return expression.ArrowFunctionExpr()

        body = [
            expression.RawFunctionExpr(self.render_emitter.emit_render(child))
            for child in vif.children
        ]

        return expression.ArrowFunctionExpr(body=expression.ListExpr.from_values(*body))



from typing import Optional
from instaui.internal import import_presets
from instaui.internal.ast.core import (
    MatchCase,
    MatchRender,
    Render,
)
from instaui.internal.ast import expression
from instaui.internal.ast.property_key import ExpressionKey
from instaui.internal.codegen.base import RenderEmitter
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.expr_codegen import ExpressionCodegen


class MatchRenderer:
    def __init__(self, ctx: FileCodegenContext, render_emitter: RenderEmitter) -> None:
        self.ctx = ctx
        self.expr = ExpressionCodegen(ctx)
        self.render_emitter = render_emitter

    def emit(self, match: MatchRender) -> str:
        method = self.ctx.imports.use_from_preset(import_presets.Instaui.match())

        cond = match.condition

        opts = expression.ObjectExpr.from_kwargs(
            cs=self.emit_cases(match.cases),
            df=self.emit_default_case(match.default_case),
        )

        return f"{method}({self.expr.emit(cond)},{self.expr.emit(opts)})"

    def emit_cases(self, cases: list[MatchCase]) -> Optional[expression.ObjectExpr]:
        if not cases:
            return None

        return expression.ObjectExpr.from_pairs(
            (
                ExpressionKey(case.value),
                expression.ArrowFunctionExpr(self.parse_renders_body(case.children)),
            )
            for case in cases
        )

    def emit_default_case(
        self, default_case: Optional[list[Render]]
    ) -> Optional[expression.ArrowFunctionExpr]:
        if not default_case:
            return None

        return expression.ArrowFunctionExpr(self.parse_renders_body(default_case))

    def parse_renders_body(self, renders: list[Render]):
        body = [
            expression.RawFunctionExpr(self.render_emitter.emit_render(child))
            for child in renders
        ]

        return expression.ListExpr.from_values(*body)



from instaui.internal.backend.common.response import response_data
from .watch_handler import get_watch_handler
from .event_handler import get_event_handler


class Api:
    def watch_call(self, data: dict):
        hkey = data.pop("hKey")
        handler_info = get_watch_handler(hkey)
        if handler_info is None:
            return {"error": "watch handler not found"}

        result = handler_info.fn(
            *handler_info.get_handler_args(_get_binds_from_data(data))
        )
        return response_data(handler_info.outputs_binding_count, result)

    def event_call(self, data: dict):
        hkey = data.pop("hKey")
        handler = get_event_handler(hkey)
        if handler is None:
            raise ValueError("event handler not found")

        args = [bind for bind in data.get("bind", [])]

        result = handler.fn(*handler.get_handler_args(args))
        return response_data(handler.outputs_binding_count, result)


def _get_binds_from_data(data: dict):
    return data.get("input", [])



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
from instaui.internal.assets import PluginDependencyInfo
from instaui.internal.assets.plugin import PluginOptionsBuilder
from instaui.internal.ast.property_key import StringKey
from instaui.internal.pipeline.ast_build.component_builder import ComponentBuilder
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder
from instaui.internal.ast import core as ast_core
from instaui.internal.ast import expression
from instaui.internal.pipeline.ast_build.context import AstBuilderContext
from instaui.internal.ui._scope import Scope

if TYPE_CHECKING:
    from instaui.internal.runtime.runtime_ctx import RuntimeContext
    from instaui.internal.runtime.render_ctx import RenderContext


class PluginsBuilder:
    def __init__(
        self,
        runtime_ctx: RuntimeContext,
        render_ctx: RenderContext,
        builder_ctx: AstBuilderContext,
    ) -> None:
        self.runtime_ctx = runtime_ctx
        self.ctx = builder_ctx
        self._expr_builder = ExprBuilder(self.ctx)
        self._component_builder = ComponentBuilder(runtime_ctx, render_ctx, self.ctx)

    def build(
        self, plugins: Sequence[PluginDependencyInfo]
    ) -> Optional[list[ast_core.Plugin]]:
        if not plugins:
            return None

        return [
            ast_core.Plugin(
                url=expression.StringLiteral(
                    self.ctx.asset_resolver.resolve_asset(plugin.esm)
                ),
                options=self._parse_options(plugin.options),
            )
            for plugin in plugins
        ]

    def _parse_options(self, options: Optional[dict | PluginOptionsBuilder]):
        if options is None:
            return None

        if isinstance(options, dict):
            return expression.JsonLiteralExpr(options)

        opts_build_fn = options

        return self._expr_builder.object_expr(
            opts_build_fn(self._parse_option_key, self._parse_scope)
        )

    def _parse_scope(self, scope: Scope):
        return expression.IdentifierRef(self.ctx.scope_mapper.get_or_create_id(scope))

    def _parse_option_key(self, key: str):
        return StringKey(key)



from instaui.internal.ast.core import (
    Render,
    ComponentRef,
)
from instaui.internal.ast.expression import Expression


class ComponentRefCollector:
    def collect(self, expr: Expression, out: set[ComponentRef]) -> None:
        if isinstance(expr, ComponentRef):
            out.add(expr)

        elif isinstance(expr, Render):
            for v in expr.props.values():
                self.collect(v, out)

        else:
            for attr in vars(expr).values():
                if isinstance(attr, Expression):
                    self.collect(attr, out)
                elif isinstance(attr, (list, tuple)):
                    for x in attr:
                        if isinstance(x, Expression):
                            self.collect(x, out)



from typing import Iterable
from instaui.internal.ast.core import App, Element
from instaui.internal.ast.expression import CallExpr, Literal, IdentifierRef, ObjectExpr
from instaui.internal.ast.sfc_template import (
    TemplateDirectiveInfo,
    TemplateElementInfo,
    TemplateTagKind,
)
from instaui.internal.pipeline.normalize.base import AstPass

SUPPORTED_VUE_DIRECTIVES: dict[str, int] = {
    "if": 1,
    "for": 2,
    "show": 1,
}


class NormalizeForVueSfcTemplate(AstPass):
    def run(self, ast: App):
        """
        In-place normalize.
        If no exception is raised, ast is guaranteed to be Vue-template-safe.
        """
        for el in self.walk_elements(ast):
            self.normalize_element_for_vue(el)

        return ast

    def walk_elements(self, ast: App) -> Iterable[Element]:
        for comp in ast.components:
            for render in comp.renders:
                if isinstance(render, Element):
                    yield render

    def normalize_element_for_vue(self, el: Element) -> None:
        tag_kind = self.normalize_tag_for_vue(el)

        props_ok = self.normalize_props_for_vue(el)
        events_ok = self.normalize_events_for_vue(el)
        directive_infos = self.normalize_directives_for_vue(el)

        el._tpl = TemplateElementInfo(
            tag_kind=tag_kind,
            props_is_object_literal=props_ok,
            events_is_object_literal=events_ok,
            directives=directive_infos,
        )

    def normalize_tag_for_vue(self, el: Element) -> TemplateTagKind:
        tag = el.tag

        if isinstance(tag, Literal):
            return TemplateTagKind.HTML

        if isinstance(tag, IdentifierRef):
            return TemplateTagKind.COMPONENT

        # <component :is="expr">
        return TemplateTagKind.DYNAMIC

    def normalize_props_for_vue(self, el: Element) -> bool:
        if el.props is None:
            return False

        if not isinstance(el.props, ObjectExpr):
            raise TemplateNormalizeError(
                "In Vue SFC template mode, element props must be an object literal"
            )

        return True

    def normalize_events_for_vue(self, el: Element) -> bool:
        if el.events is None:
            return False

        if not isinstance(el.events, ObjectExpr):
            raise TemplateNormalizeError(
                "In Vue SFC template mode, events must be an object literal"
            )

        return True

    def normalize_directives_for_vue(self, el: Element) -> list[TemplateDirectiveInfo]:
        infos: list[TemplateDirectiveInfo] = []

        if not el.dirs:
            return infos

        for call in el.dirs:
            if not isinstance(call, CallExpr):
                raise TemplateNormalizeError("Directive must be a call expression")

            if not isinstance(call.callee, IdentifierRef):
                raise TemplateNormalizeError("Directive callee must be an identifier")

            name = call.callee.name
            if name not in SUPPORTED_VUE_DIRECTIVES:
                raise TemplateNormalizeError(f"Unsupported Vue directive: {name}")

            expected_args = SUPPORTED_VUE_DIRECTIVES[name]
            actual_args = len(call.args or [])

            if actual_args != expected_args:
                raise TemplateNormalizeError(
                    f"Directive '{name}' expects {expected_args} arguments, "
                    f"got {actual_args}"
                )

            infos.append(
                TemplateDirectiveInfo(
                    name=name,
                    arg_count=expected_args,
                )
            )

        return infos


class TemplateNormalizeError(Exception):
    pass



from __future__ import annotations
from typing import (
    Any,
    Callable,
    Generic,
    Optional,
    Sequence,
    TypeVar,
    cast,
)
from typing_extensions import ParamSpec
from collections.abc import Hashable
from instaui.debug.api_boundary import user_api
from instaui.debug.model import SourceSpan
from instaui.debug.source import get_source_span
from instaui.internal.codec import build_type_adapter_map
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.protocol import ObservableProtocol
from instaui.protocol.invocation.spec import ComputedSpec

from .variable import Variable
from .bindable import BindableMixin, VarableBindHelper
from .app_context import get_app, get_current_scope
from .app_index import new_scope_if_needed
from .enums import InputBindingType
from .path_var import PathVar
from .pre_setup import normalize_pre_setup


P = ParamSpec("P")
R = TypeVar("R")


class WebComputed(
    PathVar,
    BindableMixin,
    Variable,
    ObservableProtocol,
    Generic[P, R],
):
    def __init__(
        self,
        func: Callable[P, R],
        inputs: Optional[Sequence] = None,
        extend_outputs: Optional[Sequence] = None,
        init_value: Optional[R] = None,
        evaluating: Optional[Any] = None,
        deep_compare_on_input: bool = False,
        pre_setup: Optional[list] = None,
        debug_info: Optional[dict] = None,
        extra_key: Optional[Sequence[Hashable]] = None,
        *,
        _source_span_: SourceSpan,
    ) -> None:
        self._source_span_ = _source_span_
        new_scope_if_needed()
        define_scope = get_current_scope()
        self._pre_setup = normalize_pre_setup(pre_setup or [])

        self._inputs = UILiteralExpr.try_parse_list(inputs or [])
        self._extend_outputs = extend_outputs or []
        self._fn = func
        self._init_value = UILiteralExpr.try_parse(init_value)
        self._deep_compare_on_input = deep_compare_on_input

        if evaluating is not None:
            self._pre_setup.append([evaluating, True, False])

        if debug_info is not None:
            self.debug = debug_info

        self._extra_key = extra_key

        self._bind_helper = VarableBindHelper(
            self,
            define_scope.register_web_computed,
            define_scope=define_scope,
            lazy_mark_used=[
                *self._inputs,
                *(extend_outputs or []),
                *[x[0] for x in self._pre_setup],
            ],
        )

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self._fn(*args, **kwargs)

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _export_invocation_spec(self) -> ComputedSpec:
        type_adapter_map = build_type_adapter_map(self._fn, self._inputs)

        return ComputedSpec(
            len(self._extend_outputs) + 1, custom_type_adapter_map=type_adapter_map
        )


@user_api
def web_computed(
    *,
    inputs: Optional[Sequence] = None,
    extend_outputs: Optional[Sequence] = None,
    init_value: Optional[Any] = None,
    evaluating: Optional[Any] = None,
    deep_compare_on_input: bool = False,
    pre_setup: Optional[list] = None,
    extra_key: Optional[Sequence[Hashable]] = None,
    debug_info: Optional[dict] = None,
):
    """
    Creates a computed property decorator for reactive programming with dependency tracking.

    This decorator factory wraps functions to create reactive computed properties that:
    - Automatically re-evaluate when dependencies (inputs) change
    - Cache results for performance optimization
    - Support both synchronous and asynchronous computation patterns

    Args:
        inputs (Optional[Sequence], optional): Collection of reactive sources that trigger recomputation
                                   when changed. These can be state objects or other computed properties.
        extend_outputs (Optional[Sequence], optional):  Additional outputs to notify when this computed value updates.
        init_value (Optional[Any], optional): Initial value to return before first successful evaluation.
        evaluating (Optional[Any], optional): Temporary value returned during asynchronous computation.
        pre_setup (typing.Optional[list], optional): A list of pre-setup actions to be executed before the event executes.

    # Example:
    .. code-block:: python
        from instaui import ui,html

        a = ui.state(0)

        @ui.computed(inputs=[a])
        def plus_one(a):
            return a + 1

        html.number(a)
        ui.text(plus_one)
    """

    if get_app().mode == "zero":
        raise Exception(
            "Cannot use computed decorator in zero mode. You should use `ui.js_computed` instead."
        )

    _source_span_ = get_source_span()

    def wrapper(func: Callable[P, R]):
        return cast(
            R,
            WebComputed(
                func,
                inputs=inputs,
                extend_outputs=extend_outputs,
                init_value=init_value,
                evaluating=evaluating,
                deep_compare_on_input=deep_compare_on_input,
                pre_setup=pre_setup,
                debug_info=debug_info,
                extra_key=extra_key,
                _source_span_=_source_span_,
            ),
        )

    return wrapper


TComputed = WebComputed



from typing import Any, Mapping, Optional
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.app_index import new_scope_if_needed
from instaui.internal.ui.bindable import BindableMixin, VarableBindHelper
from instaui.internal.ui.enums import InputBindingType
from instaui.internal.ui.path_var import PathVar
from instaui.internal.ui.protocol import ObservableProtocol
from instaui.internal.ui.variable import Variable


class ExprComputed(Variable, PathVar, BindableMixin, ObservableProtocol):
    def __init__(
        self,
        fn_code: str,
        bindings: Optional[Mapping[str, Any]] = None,
    ) -> None:
        new_scope_if_needed()
        define_scope = get_current_scope()
        self.code = fn_code
        self._bindings = UILiteralExpr.try_parse_dict(bindings) if bindings else None

        self._bind_helper = VarableBindHelper(
            self,
            define_scope.register_expr_computed,
            define_scope=define_scope,
            lazy_mark_used=[self._bindings],
        )

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref


TExprComputed = ExprComputed



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional


from .app_context import get_app

if TYPE_CHECKING:
    from .renderable import Renderable
    from instaui.internal.ui._scope import Scope


class Container(ABC):
    @abstractmethod
    def add_child(self, renderable: Renderable): ...

    def __enter__(self):
        get_app().push_container(self)
        return self

    def __exit__(self, *_):
        self._release_scope()
        get_app().pop_container()

    def _bind_scope(self, scope: Scope): ...

    def _release_scope(self): ...


class ScopeGuard:
    def __init__(self):
        self.scope: Optional[Scope] = None

    def release_scope(self):
        if self.scope is None:
            return
        app = get_app()
        app.pop_scope()
        app.pop_container()
        self.scope = None

    def bind_scope(self, scope: Scope):
        self.scope = scope



from typing import Optional, Union, cast
from instaui.constants.ui import DEFAULT_SLOT_NAME
from instaui.internal import import_presets
from instaui.internal.ast import value_rewriter
from instaui.internal.ast.core import (
    Directive,
    Element,
    ElementClasses,
    ElementProp,
    ElementStyles,
    Slot,
)
from instaui.internal.ast import expression
from instaui.internal.ast.property_key import StringKey
from instaui.internal.codegen.base import RenderEmitter
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.expr_codegen import ExpressionCodegen


class ElementRenderer:
    def __init__(self, ctx: FileCodegenContext, render_emitter: RenderEmitter) -> None:
        self.ctx = ctx
        self.expr = ExpressionCodegen(ctx)
        self.render_emitter = render_emitter

    def emit(self, element: Element) -> str:
        method = self.ctx.imports.use_from_preset(
            import_presets.Instaui.render_component()
        )

        obj = expression.ObjectExpr.from_kwargs(
            tag=element.tag,
            props=self.normalize_props(element.props),
            classes=self.normalize_classes(element.classes),
            style=self.normalize_styles(element.styles),
            events=self.normalize_events(element.events),
            dirs=self.normalize_directives(element.dirs),
            lifeEvents=self.normalize_life_events(element.lifecycles),
            slots=self.emit_slots(element),
        )

        return f"{method}({self.expr.emit(obj)})"

    def emit_slots(
        self, element: Element
    ) -> Union[expression.ObjectExpr, expression.ArrowFunctionExpr, None]:
        if not element.slots:
            return None

        slots: dict[str, expression.Expression] = {}

        for slot in element.slots.values():
            slots[self.expr.emit(slot.name)] = self.emit_slot_body(slot)

        if len(slots) == 1 and DEFAULT_SLOT_NAME in slots:
            return cast(expression.ArrowFunctionExpr, slots[DEFAULT_SLOT_NAME])

        return expression.ObjectExpr.from_kwargs(**slots)

    def emit_slot_body(self, slot: Slot) -> expression.Expression:
        exprs = [
            expression.RawFunctionExpr(self.render_emitter.emit_render(n))
            for n in slot.body
        ]

        params = expression.IdentifierRef(slot.used_prop) if slot.used_prop else None

        return expression.ArrowFunctionExpr(
            body=expression.ListExpr.from_values(*exprs), params=params
        )

    def normalize_props(
        self, props: Optional[ElementProp]
    ) -> Optional[expression.ObjectExpr]:
        if not props:
            return None

        return expression.ObjectExpr.from_kwargs(
            bProps=expression.ObjectExpr.from_dict(
                value_rewriter.rewrite_string_key_shallow(props.binding)
            )
            if props.binding
            else None,
            pProps=expression.ListExpr.from_values(*props.proxy)
            if props.proxy
            else None,
            sProps=props.static,
            ref=props.ref,
        )

    def normalize_life_events(
        self, lifecycles: Optional[dict[str, expression.Expression]]
    ) -> Optional[expression.ObjectExpr]:
        if not lifecycles:
            return None

        return expression.ObjectExpr.from_dict(
            {event_name: event_args for event_name, event_args in lifecycles.items()}
        )

    def normalize_styles(
        self, styles: Optional[ElementStyles]
    ) -> Optional[expression.ObjectExpr]:
        if not styles:
            return None

        return expression.ObjectExpr.from_kwargs(
            sStyle=styles.static if styles.static else None,
            bStyle=styles.binding if styles.binding else None,
            pStyle=expression.ListExpr.from_values(*styles.proxy)
            if styles.proxy
            else None,
        )

    def normalize_classes(
        self, classes: Optional[ElementClasses]
    ) -> Optional[expression.Expression]:
        if not classes:
            return None
        if classes.static and not classes.binding and not classes.maps:
            return expression.Literal(" ".join(classes.static.value))

        return expression.ObjectExpr.from_kwargs(
            sClass=classes.static if classes.static else None,
            bClass=expression.ListExpr.from_values(*classes.binding)
            if classes.binding
            else None,
            mClass=classes.maps if classes.maps else None,
        )

    def normalize_directives(self, dirs: Optional[list[Directive]]):
        if not dirs:
            return None

        return expression.ListExpr.from_values(
            *[
                expression.ObjectExpr.from_kwargs(
                    name=dir.name,
                    value=dir.value,
                    sys=dir.sys,
                    arg=dir.arg,
                    mf=dir.mf,
                )
                for dir in dirs
            ]
        )

    def normalize_events(
        self, arg: Optional[expression.ObjectExpr]
    ) -> Optional[expression.ObjectExpr]:
        if not arg:
            return None
        return expression.ObjectExpr.from_pairs(
            (_normalize_event_name(prop), prop.value) for prop in arg.props
        )


def _normalize_event_name(prop: expression.ObjectProperty) -> StringKey:
    """'click' -> 'onClick' , 'press-enter' -> 'onPressEnter' , 'pressEnter' -> 'onPressEnter'"""

    event_name = cast(StringKey, prop.key).value

    if event_name.startswith("on-"):
        event_name = event_name[3:]

    if event_name.startswith("on"):
        event_name = event_name[2:]

    parts = event_name.split("-")
    formatted_parts = [part[0].upper() + part[1:] for part in parts]

    return StringKey("".join(["on", *formatted_parts]))



from instaui.internal.pipeline.ast_build.base import RenderBuilderProtocol
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder

from instaui.internal.ui.vfor import VFor
from instaui.internal.ast import core as ast_core
from instaui.internal.ast import expression
from instaui.internal.pipeline.ast_build.context import AstBuilderContext


class VForParser:
    def __init__(
        self, ctx: AstBuilderContext, render_builder: RenderBuilderProtocol
    ) -> None:
        self.ctx = ctx
        self._expr_builder = ExprBuilder(ctx)
        self._render_builder = render_builder

    def parse(self, vfor: VFor) -> ast_core.VFor:
        key = expression.Literal(vfor._key)

        if vfor._data:
            value = self._expr_builder.build(vfor._data, mode="literal")
        elif vfor._range:
            value = self._expr_builder.object_expr(vfor._range)
        else:
            value = expression.UNDEFINED

        array = ast_core.VForArray(
            expression.Literal(vfor._array_type.value),
            value,
        )

        used_item = None
        used_index = None
        used_key = None

        if vfor._vfor_item._used:
            used_item = self.ctx.var_mapper.get_or_create_id(vfor._vfor_item)
            self.ctx.component_scope_manager.declare_var(used_item)

        if vfor._vfor_index._used:
            used_index = self.ctx.var_mapper.get_or_create_id(vfor._vfor_index)
            self.ctx.component_scope_manager.declare_var(used_index)

        if vfor._vfor_item_key._used:
            used_key = self.ctx.var_mapper.get_or_create_id(vfor._vfor_item_key)
            self.ctx.component_scope_manager.declare_var(used_key)

        return ast_core.VFor(
            key,
            array,
            used_item=used_item,
            used_index=used_index,
            used_key=used_key,
            transition_group=expression.ObjectLiteral(vfor._transition_group_setting)
            if vfor._transition_group_setting
            else None,
            children=[self._render_builder.build(r) for r in vfor._renderables],
        )



import re
from .base import RouteSyntaxAdapter


_FASTAPI_PARAM_RE = re.compile(
    r"{\s*(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)\s*(?::[a-zA-Z_]+)?\s*}"
)


class FastAPIRouteSyntaxAdapter(RouteSyntaxAdapter):
    """
    Normalize FastAPI-style route syntax into frontend-friendly form.

    Examples:
        "/users/{id}"        -> "/users/:id"
        "/users/{id:int}"    -> "/users/:id"
        "/files/{path:path}" -> "/files/:path"
    """

    def normalize(self, path: str) -> str:
        return _FASTAPI_PARAM_RE.sub(
            lambda m: f":{m.group('name')}",
            path,
        )



from typing import Callable, Mapping, Optional
from instaui.internal.backend.common import handler as _handler
from instaui.internal.backend.common.event_dataset import event_dataset_type_adapter
from instaui.protocol.codec.type_adapter import TypeAdapterProtocol


_event_handlers: dict[str, _handler.HandlerInfo] = {}


def register_event_handler(
    key: str,
    handler: Callable,
    outputs_binding_count: int,
    dataset_input_indexs: list[int],
    custom_type_adapter_map: Optional[Mapping[int, TypeAdapterProtocol]] = None,
):
    if key in _event_handlers:
        return

    custom_type_adapter_map = dict(custom_type_adapter_map or {}) | {
        i: event_dataset_type_adapter for i in dataset_input_indexs
    }

    _event_handlers[key] = _handler.HandlerInfo.from_handler(
        handler,
        outputs_binding_count,
        custom_type_adapter_map=custom_type_adapter_map,
    )


def get_event_handler(key: str):
    return _event_handlers.get(key)



from __future__ import annotations
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from instaui.internal.ui.element import Element


class ComponentExtensionRegistry:
    """
    Record: Additional 'component-level extension dependencies' that need to be enabled for a specific component family (target class) under a certain runtime.
    """

    def __init__(self):
        # target_cls -> kind -> values
        self._by_target: dict[type[Element], dict[str, set[str]]] = {}

    def add(
        self,
        *,
        target: type["Element"],
        kind: str,
        values: list[str],
    ):
        self._by_target.setdefault(target, {}).setdefault(kind, set()).update(values)

    def items(self):
        return self._by_target.items()

    def overridden_by(
        self, other: ComponentExtensionRegistry
    ) -> ComponentExtensionRegistry:
        """
        Return a NEW registry where targets in `other`
        completely replace those in `self`.

        Example:
            global_registry = {'t1': v1, 't3': v3}
            local_registry = {'t1': v2}

            global_registry.overridden_by(local_registry)
            # {'t1': v2, 't3': v3}
        """
        result = ComponentExtensionRegistry()

        for target, kinds in self._by_target.items():
            if target in other._by_target:
                continue

            for kind, values in kinds.items():
                result.add(
                    target=target,
                    kind=kind,
                    values=list(values),
                )

        for target, kinds in other._by_target.items():
            for kind, values in kinds.items():
                result.add(
                    target=target,
                    kind=kind,
                    values=list(values),
                )

        return result



from typing import TypeVar, cast
from instaui.constants.ui import (
    USE_ROUTE_PARAM_METHOD,
    USE_QUERY_PARAM_METHOD,
    INSTAUI_JS_MODULE_NAME,
)
from instaui.internal.ui.custom_var import CustomVar

_T = TypeVar("_T")


def route_param(name: str, *, default: _T = None) -> _T:
    return cast(
        _T,
        CustomVar(INSTAUI_JS_MODULE_NAME, USE_ROUTE_PARAM_METHOD, args=(name, default)),
    )


def query_param(name: str, *, default: _T = None) -> _T:
    return cast(
        _T,
        CustomVar(INSTAUI_JS_MODULE_NAME, USE_QUERY_PARAM_METHOD, args=(name, default)),
    )



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Protocol, Union
from instaui.systems.dataclass_system import dataclass, field

if TYPE_CHECKING:
    from instaui.internal.ui._scope import Scope
    from instaui.internal.assets.css_assets import CssAsset


class ParseScopeFn(Protocol):
    def __call__(self, scope: Scope) -> Any: ...


class StringKeyFn(Protocol):
    def __call__(self, key: str) -> Any: ...


class PluginOptionsBuilder(Protocol):
    def __call__(self, parse_key: StringKeyFn, parse_scope: ParseScopeFn) -> dict: ...


@dataclass(frozen=True)
class PluginDependencyInfo:
    name: str = field(hash=True)
    esm: Path = field(hash=False)
    externals: dict[str, Path] = field(default_factory=dict, compare=False, hash=False)
    css: list[Union[str, Path, CssAsset]] = field(
        default_factory=list, compare=False, hash=False
    )
    options: Optional[dict | PluginOptionsBuilder] = field(hash=False, default=None)



from dataclasses import dataclass
from typing import List
from instaui.internal.ast.core import ComponentDef


@dataclass
class FileUnit:
    path: str  # "components/Foo.js"
    components: List[ComponentDef]
    is_entry: bool = False



from __future__ import annotations
from typing import TYPE_CHECKING, Protocol


if TYPE_CHECKING:
    from instaui.internal.ast.core import Render


class RenderEmitter(Protocol):
    def emit_render(self, render: Render) -> str: ...



from instaui.systems.dataclass_system import dataclass


@dataclass(frozen=True)
class SourceSpan:
    file: str
    line: int
    function: str | None
    column: int | None = None

    def __str__(self) -> str:
        if self.column is not None:
            loc = f"{self.file}:{self.line}:{self.column}"
        else:
            loc = f"{self.file}:{self.line}"

        if self.function:
            return f"{loc} in {self.function}()"
        return loc



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING, Mapping, Optional, Sequence
from instaui.constants.runtime import RuntimeMode
from instaui.systems.dataclass_system import dataclass, field


if TYPE_CHECKING:
    from instaui.internal.assets.css_assets import CssAsset
    from instaui.internal.ui.element import Element


@dataclass(frozen=True)
class ComponentDependencyOverride:
    zero_externals: Mapping[str, Path] | None = None
    zero_css: Sequence[str | Path | CssAsset] | None = None


@dataclass(frozen=True)
class ComponentDependencyInfo:
    tag_name: str = field(hash=True)
    esm: Path = field(hash=False)
    externals: dict[str, Path] = field(default_factory=dict, compare=False, hash=False)
    css: list[str | Path | CssAsset] = field(
        default_factory=list, compare=False, hash=False
    )
    overrides: Optional[Mapping[RuntimeMode, ComponentDependencyOverride]] = field(
        default=None, compare=False, hash=False
    )


@dataclass(frozen=True)
class ComponentDependencyRecord:
    component: type[Element]
    dependency: ComponentDependencyInfo


class ComponentDependencyRegistry:
    def __init__(self):
        self._records: set[ComponentDependencyRecord] = set()

    def add(self, component: type[Element], dep: ComponentDependencyInfo):
        self._records.add(ComponentDependencyRecord(component, dep))

    @property
    def records(self) -> set[ComponentDependencyRecord]:
        return self._records



from __future__ import annotations
from typing import Iterable

from instaui.systems.dataclass_system import dataclass
from instaui.internal.ast.nodes import ImportMember, ImportStatement


class ModuleImportCodegen:
    def emit_members(self, members: Iterable[ResolvedImportMember]) -> str:
        return ", ".join(
            f"{m.name} as {m.alias}" if m.alias else m.name for m in members
        )

    def emit_module_import(self, statement: ResolvedImportStatement) -> str:
        return (
            f"import {{{self.emit_members(statement.members)}}} "
            f"from '{statement.module_name}';"
        )

    def emit_code(
        self, import_statements: Iterable[ImportStatement]
    ) -> ModuleImportResult:
        merged = self._dedupe_and_merge(import_statements)

        used_names: dict[str, int] = {}
        names_map: dict[MemberSource, str] = {}
        resolved_statements: list[ResolvedImportStatement] = []

        for module_name, members in merged.items():
            resolved_members: list[ResolvedImportMember] = []

            for member in members:
                public_name = member.alias or member.name
                source = MemberSource(module_name, member.name)

                if public_name not in used_names:
                    used_names[public_name] = 0
                    final_name = public_name
                    alias = member.alias
                else:
                    used_names[public_name] += 1
                    suffix = used_names[public_name]
                    final_name = f"_{public_name}{suffix}"
                    alias = final_name

                names_map[source] = final_name

                resolved_members.append(
                    ResolvedImportMember(
                        name=member.name,
                        alias=alias,
                    )
                )

            resolved_statements.append(
                ResolvedImportStatement(
                    module_name=module_name,
                    members=resolved_members,
                )
            )

        code = "\n".join(self.emit_module_import(stmt) for stmt in resolved_statements)

        return ModuleImportResult(code, names_map)

    def _dedupe_and_merge(
        self, import_statements: Iterable[ImportStatement]
    ) -> dict[str, list[ImportMember]]:
        """
        Returns a dictionary of module names to lists of import members.
        {
            module_name: [ImportMember, ...]
        }
        """
        modules: dict[str, dict[str, ImportMember]] = {}

        for stmt in import_statements:
            module_members = modules.setdefault(stmt.module_name, {})

            for member in stmt.members:
                # member.name is unique within a module
                if member.name not in module_members:
                    module_members[member.name] = member

        return {module: list(members.values()) for module, members in modules.items()}


# =========================
# Resolved-only structures
# =========================


@dataclass(frozen=True)
class ResolvedImportMember:
    name: str
    alias: str | None = None


@dataclass(frozen=True)
class ResolvedImportStatement:
    module_name: str
    members: list[ResolvedImportMember]


@dataclass()
class ModuleImportResult:
    code: str
    names_map: dict[MemberSource, str]


@dataclass(frozen=True)
class MemberSource:
    module_name: str
    member_name: str



from enum import Enum
from typing import Any, Callable, Optional, Protocol, Type
from typing_extensions import dataclass_transform
import sys
from dataclasses import (
    dataclass as _dataclass,
    fields,
    Field,
    field as _field,
    asdict as _asdict,
    is_dataclass,
    replace as _replace,
)


PY_VERSION = sys.version_info


field = _field
asdict = _asdict
replace = _replace


@dataclass_transform(field_specifiers=(field,))
def dataclass(
    _cls: Optional[Type[Any]] = None, *, slots: bool = True, **kwargs: Any
) -> Callable[[Type[Any]], Type[Any]]:
    """
    A thin wrapper around dataclasses.dataclass for Python 3.10+.

    Features:
    - slots=True by default
    - Fully equivalent to @dataclass(slots=...)
    - Works with IDE/mypy type hints for smart __init__ completion
    """

    def wrap(cls: Type[Any]) -> Type[Any]:
        return _dataclass(cls, slots=slots, **kwargs)

    if _cls is None:
        # @dataclass(...) usage
        return wrap
    # @dataclass usage
    return wrap(_cls)


class KeyResolver(Protocol):
    def __call__(self, field: Field) -> str: ...


def default_key_resolver(f) -> str:
    return f.name


def metadata_key_resolver(meta_key: str = "key"):
    def _resolver(f) -> str:
        return f.metadata.get(meta_key, f.name)

    return _resolver


def asdict_with_alias(
    obj: Any,
    *,
    key_resolver=default_key_resolver,
) -> Any:
    if is_dataclass(obj):
        result = {}
        for f in fields(obj):
            key = key_resolver(f)
            value = getattr(obj, f.name)
            result[key] = asdict_with_alias(value, key_resolver=key_resolver)
        return result

    if isinstance(obj, list):
        return [asdict_with_alias(v, key_resolver=key_resolver) for v in obj]

    if isinstance(obj, tuple):
        return tuple(asdict_with_alias(v, key_resolver=key_resolver) for v in obj)

    if isinstance(obj, dict):
        return {
            k: asdict_with_alias(v, key_resolver=key_resolver) for k, v in obj.items()
        }

    return obj


def asdict_no_none(obj, *, key_resolver=default_key_resolver) -> dict:
    def _clean(v: Any):
        if isinstance(v, Enum):
            return v.value

        if is_dataclass(v):
            return _clean(asdict(v))  # type: ignore
        if isinstance(v, dict):
            return {k: _clean(vv) for k, vv in v.items() if vv is not None}
        if isinstance(v, list):
            return [_clean(i) for i in v]
        return v

    return _clean(asdict_with_alias(obj, key_resolver=key_resolver))  # type: ignore



from instaui.debug.api_boundary import user_api
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.app_index import new_scope_if_needed
from instaui.internal.ui.bindable import BindableMixin, VarableBindHelper
from instaui.internal.ui.enums import InputBindingType
from instaui.internal.ui.protocol import CanInputProtocol
from instaui.internal.ui.variable import Variable
from instaui.debug.source import get_source_span


class JsFn(
    Variable,
    CanInputProtocol,
    BindableMixin,
):
    def __init__(self, code: str, *, execute_immediately=False):
        self._source_span_ = get_source_span()
        new_scope_if_needed()

        self.code = code
        self._execute_immediately = execute_immediately

        define_scope = get_current_scope()
        self._bind_helper = VarableBindHelper(
            self, define_scope.register_js_fn, define_scope=define_scope
        )

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.JsFn


@user_api
def js_fn(code: str, *, execute_immediately=False):
    """
    Creates a JavaScript function object from a raw code string.
    Valid targets include: `js_computed`, `js_watch`, and similar JS-bound methods.

    Args:
        code (str): Valid JavaScript function definition string.

    Example:
    .. code-block:: python
        a = ui.state(1)
        add = ui.js_fn(code="(a,b)=> a+b ")
        result = ui.js_computed(inputs=[add, a], code="(add,a)=>  add(a,10) ")

        html.number(a)
        ui.text(result)
    """

    return JsFn(code, execute_immediately=execute_immediately)



from pathlib import Path
from typing import Union
from instaui.systems.dataclass_system import dataclass


@dataclass(frozen=True)
class WebUserLaunchConfig:
    caller_folder_path: Path
    debug: bool
    use_gzip: Union[int, bool] = True
    prefix: str = ""
    prerender: bool = False
    prerender_concurrency: int = 4



from __future__ import annotations
from typing import Optional
from instaui.internal.ui.element import Element
from instaui.internal.ui.value_element import ValueElement
from ._mixins import InputEventMixin


_T_value = str


class Radio(InputEventMixin, ValueElement[_T_value]):
    def __init__(
        self,
        value: Optional[_T_value] = None,
        *,
        model_value: Optional[_T_value] = None,
        id: Optional[str] = None,
        name: Optional[str] = None,
    ):
        super().__init__("input", value, is_html_component=True)
        self.props({"type": "radio"})

        self.props(
            {
                "id": id,
                "name": name,
            }
        )

        if model_value is not None:
            self.props({"value": model_value})

    def _input_event_mixin_element(self) -> Element:
        return self



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeVar, Callable


from instaui.internal.codegen.components.protocol import (
    ImportTableProtocol,
    AppBootstrapCodegenProtocol,
)

if TYPE_CHECKING:
    from instaui.internal.codegen.context.file_ctx import FileCodegenContext

T = TypeVar("T")


@dataclass
class CodegenComponents:
    """
    Describing 'How to Construct Various Components Required for Codegen'
    """

    import_table_factory: Callable[[], ImportTableProtocol]
    app_bootstrap_codegen_factory: Callable[
        [FileCodegenContext], AppBootstrapCodegenProtocol
    ]

    @classmethod
    def default(cls) -> CodegenComponents:
        from instaui.internal.codegen.import_table import ImportTable
        from instaui.internal.codegen.app_bootstrap_codegen import AppBootstrapCodegen

        return cls(
            import_table_factory=ImportTable,
            app_bootstrap_codegen_factory=AppBootstrapCodegen,
        )



from .model import PageInfo


class RouteRegistry:
    def __init__(self):
        self.routes: dict[str, PageInfo] = {}  # path -> function

    def add_route(self, path: str, info: PageInfo):
        self.routes[path] = info


route_registry = RouteRegistry()



from pathlib import Path
from typing import Mapping
from instaui.constants.runtime import RuntimeMode
from instaui.internal.assets.base import AssetsDeclaration
from instaui.internal.assets.component_dep import ComponentDependencyInfo


class ComponentImportMapProcessor:
    def process(self, assets: AssetsDeclaration) -> None:
        for comp in assets.component_dependencies.records:
            for k, v in comp.dependency.externals.items():
                assets.import_maps.setdefault(k, v)


class ZeroModeComponentImportMapProcessor:
    def process(self, assets: AssetsDeclaration) -> None:
        for record in assets.component_dependencies.records:
            dependency = record.dependency

            externals = self._resolve_zero_mode_externals(dependency)

            for key, path in externals.items():
                assets.import_maps.setdefault(key, path)

    @staticmethod
    def _resolve_zero_mode_externals(
        dependency: ComponentDependencyInfo,
    ) -> Mapping[str, Path]:
        overrides = dependency.overrides
        if not overrides:
            return dependency.externals

        zero_override = overrides.get(RuntimeMode.ZERO)
        if not zero_override or not zero_override.zero_externals:
            return dependency.externals

        return zero_override.zero_externals



import typing
from instaui.debug.api_boundary import user_api
from instaui.debug.source import get_source_span
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.app_index import new_scope_if_needed
from instaui.internal.ui.bindable import BindableMixin, VarableBindHelper, mark_used
from instaui.internal.ui.event import EventMixin
from instaui.internal.ui.event_modifier import TEventModifier
from instaui.internal.ui.variable import Variable


class JsEvent(Variable, EventMixin, BindableMixin):
    def __init__(
        self,
        code: str,
        inputs: typing.Optional[typing.Sequence] = None,
        outputs: typing.Optional[typing.Sequence] = None,
    ):
        self._source_span_ = get_source_span()
        new_scope_if_needed()
        define_scope = get_current_scope()

        self._inputs = UILiteralExpr.try_parse_list(inputs or [])
        self._outputs = outputs or []
        self._code = code

        self._bind_helper = VarableBindHelper(
            self,
            define_scope.register_js_event,
            define_scope=define_scope,
            lazy_mark_used=[*self._inputs, *(outputs or [])],
        )

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _attach_to_element(
        self,
        *,
        extends: typing.Optional[typing.Sequence],
        modifier: typing.Optional[typing.Sequence[TEventModifier]] = None,
    ):
        mark_used(extends)
        mark_used(modifier)
        self._mark_used()

        return self


@user_api
def js_event(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    outputs: typing.Optional[typing.Sequence] = None,
    code: str,
):
    """
    Creates a client-side event handler decorator for binding JavaScript logic to UI component events.

    Args:
        inputs (typing.Optional[typing.Sequence], optional):Reactive sources (state variables, computed values)
                                   that should be passed to the event handler. These values
                                   will be available in the JavaScript context through the `args` array.
        outputs (typing.Optional[typing.Sequence], optional): Targets (state variables, UI elements) that should
                                    update when this handler executes. Used for coordinating
                                    interface updates after the event is processed.
        code (str): JavaScript code to execute when the event is triggered.

    # Example:
    .. code-block:: python
        from instaui import ui, html

        a = ui.state(0)

        plus_one = ui.js_event(inputs=[a], outputs=[a], code="a =>a + 1")

        html.button("click me").on_click(plus_one)
        html.paragraph(a)

    """
    return JsEvent(inputs=inputs, outputs=outputs, code=code)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from instaui.internal.ast import expression
from instaui.internal.backend.invocation import BackendInvocation, BackendInvocationKind
from instaui.internal.pipeline.ast_build.exception import AstBuildError
from instaui.internal.pipeline.ast_build.pre_setup_parser import PreSetupParser
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder
from instaui.internal.pipeline.ast_build.utils import ui_bindings_utils

from instaui.internal.ui._scope import Scope
from instaui.internal.ast import core as ast_core
from instaui.internal.pipeline.ast_build.context import AstBuilderContext
from instaui.internal.ui.enums import OutputSetType
from instaui.internal.ui.web_computed import WebComputed
from instaui.internal.ui.web_watch import WebWatch

if TYPE_CHECKING:
    from instaui.internal.runtime.runtime_ctx import RuntimeContext
    from instaui.internal.runtime.render_ctx import RenderContext


class WebWatchTaskParser:
    def __init__(
        self,
        runtime_ctx: RuntimeContext,
        render_ctx: RenderContext,
        builder_ctx: AstBuilderContext,
    ) -> None:
        self.runtime_ctx = runtime_ctx
        self.render_ctx = render_ctx
        self._expr_builder = ExprBuilder(builder_ctx)
        self.pre_setup_parser = PreSetupParser(self._expr_builder)

    def parse(self, scope: Scope) -> Optional[list[ast_core.WebWatchTask]]:
        result: list[ast_core.WebWatchTask] = []
        if scope._web_computeds:
            result.extend(
                self._from_computed(computed) for computed in scope._web_computeds
            )

        if scope._web_watchs:
            result.extend(self._from_watch(watch) for watch in scope._web_watchs)

        return result or None

    def _from_computed(self, computed: WebComputed) -> ast_core.WebWatchTask:
        try:
            f_type = ui_bindings_utils.py_function_type(computed._fn)
            f_type = ui_bindings_utils.normalize_f_type(f_type)

            outputs = [computed, *computed._extend_outputs]

            inputs, slients, datas = ui_bindings_utils.analyze_observable_inputs(
                computed._inputs
            )

            slients = ui_bindings_utils.normalize_int_values(slients)
            datas = ui_bindings_utils.normalize_int_values(datas)

            output_types = ui_bindings_utils.normalize_int_values(
                [
                    OutputSetType.Ref.value,
                    *[x._to_event_output_type().value for x in outputs[1:]],
                ]
            )

            pre_setup = self.pre_setup_parser.parse(computed._pre_setup)

            ref_id = self.runtime_ctx.binding_registry.register(
                BackendInvocation(
                    kind=BackendInvocationKind.COMPUTED,
                    fn=computed._fn,
                    spec=computed._export_invocation_spec(),
                    render_ctx=self.render_ctx,
                )
            )

            fn = ast_core.BackendFunctionRef(ref_id)
            args = expression.ObjectExpr.from_kwargs(
                type=expression.Literal("c"),
                inputs=self._expr_builder.list_expr(mode="literal", values=inputs)
                if inputs
                else None,
                outputs=self._expr_builder.list_expr(mode="literal", values=outputs)
                if outputs
                else None,
                opTypes=expression.ListLiteral(list(output_types))
                if output_types
                else None,
                slient=expression.ListLiteral(list(slients)) if slients else None,
                data=expression.ListLiteral(list(datas)) if datas else None,
                fType=expression.Literal(f_type) if f_type else None,
                preSetup=pre_setup,
            )

            return ast_core.WebWatchTask(computed._source_span_, fn, args)
        except Exception as e:
            raise AstBuildError(
                f"Failed to build {computed}: {e}", computed._source_span_
            )

    def _from_watch(self, watch: WebWatch) -> ast_core.WebWatchTask:
        try:
            f_type = ui_bindings_utils.py_function_type(watch._fn)
            f_type = ui_bindings_utils.normalize_f_type(f_type)
            outputs = list(watch._outputs or [])
            inputs = ui_bindings_utils.auto_made_inputs_to_slient(
                list(watch._inputs or []), outputs
            )
            inputs, slients, datas = ui_bindings_utils.analyze_observable_inputs(inputs)
            slients = ui_bindings_utils.normalize_int_values(slients)
            datas = ui_bindings_utils.normalize_int_values(datas)

            output_types = ui_bindings_utils.normalize_int_values(
                [x._to_event_output_type().value for x in outputs]
            )

            pre_setup = self.pre_setup_parser.parse(watch._pre_setup)

            ref_id = self.runtime_ctx.binding_registry.register(
                BackendInvocation(
                    kind=BackendInvocationKind.WATCH,
                    fn=watch._fn,
                    spec=watch._export_invocation_spec(),
                    render_ctx=self.render_ctx,
                )
            )

            fn = ast_core.BackendFunctionRef(ref_id)
            args = expression.ObjectExpr.from_kwargs(
                type=expression.Literal("w"),
                inputs=self._expr_builder.list_expr(mode="literal", values=inputs)
                if inputs
                else None,
                outputs=self._expr_builder.list_expr(mode="literal", values=outputs)
                if outputs
                else None,
                opTypes=expression.ListLiteral(list(output_types))
                if output_types
                else None,
                slient=expression.ListLiteral(list(slients)) if slients else None,
                data=expression.ListLiteral(list(datas)) if datas else None,
                fType=expression.Literal(f_type) if f_type else None,
                preSetup=pre_setup,
                immediate=None
                if watch._immediate
                else expression.Literal(watch._immediate),
                deep=None if watch._deep else expression.Literal(watch._deep),
                once=expression.Literal(watch._once) if watch._once else None,
                flush=expression.Literal(watch._flush) if watch._flush else None,
            )

            return ast_core.WebWatchTask(watch._source_span_, fn, args)
        except Exception as e:
            raise AstBuildError(f"Failed to build {watch}: {e}", watch._source_span_)



from typing import cast

from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from .bindable import VarableBindHelper, BindableMixin
from .app_context import get_current_scope


class StrFormat(BindableMixin):
    def __init__(self, template: str, *args, **kwargs) -> None:
        self.template = UILiteralExpr.try_parse(template)
        self.args = list(UILiteralExpr.try_parse(arg) for arg in args)
        self.kwargs = {
            key: UILiteralExpr.try_parse(value) for key, value in kwargs.items()
        }

        self._bind_helper = VarableBindHelper(
            self,
            define_scope=get_current_scope(),
            lazy_mark_used=[
                template,
                args,
                kwargs,
            ],
        )

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()


def str_format(template: str, *args, **kwargs) -> str:
    """
    Formats a string template with positional and keyword arguments.

    Args:
        template (str): The string template containing {} placeholders for formatting.
        *args: Variable length positional arguments for positional formatting.
        **kwargs: Arbitrary keyword arguments for named formatting.

    Example:
    .. code-block:: python
        # Positional formatting
        ui.str_format("pos:a={},b={}", a, b)

        # Index-based positional formatting
        ui.str_format("num pos:a={0},b={1}", a, b)

        # Named formatting
        ui.str_format("name pos:b={b},a={a}", a=a, b=b)
    """
    return cast(str, StrFormat(template, *args, **kwargs))



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from instaui.systems.dataclass_system import dataclass
from instaui.internal.ui.app_context import get_current_scope

if TYPE_CHECKING:
    from instaui.internal.ui._scope import Scope


@dataclass(frozen=True)
class JSImport:
    module: str
    members: list[str]


def import_from(module: str, members: list[str], *, scope: Optional[Scope] = None):
    scope = scope or get_current_scope()
    scope.register_import(JSImport(module, members))



import base64
from pathlib import Path
from typing import Union
from urllib.parse import quote
from enum import StrEnum


class PREFIX(StrEnum):
    JS = "data:text/javascript;charset=utf-8"
    CSS = "data:text/css;charset=utf-8"
    ICON = "data:image/x-icon;base64"


def normalize_path_to_dataurl_or_cdn(path: Union[str, Path], prefix: PREFIX):
    if isinstance(path, Path):
        path = path.read_text(encoding="utf-8")
        return f"{prefix},{quote(path)}"

    return path


def normalize_path_to_base64_url(path: Path, prefix: PREFIX):
    return f"{prefix},{base64.b64encode(path.read_bytes()).decode('utf-8')}"



from __future__ import annotations
from typing import Any, Union
from instaui.internal.ui.element import Element


class Span(Element):
    def __init__(
        self,
        text: Union[str, Any],
    ):
        super().__init__("span")
        self.props(
            {
                "innerText": text,
            }
        )



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from instaui.internal.ui.enums import InputBindingType, OutputSetType


class CanInputProtocol(ABC):
    @abstractmethod
    def _to_event_input_type(self) -> InputBindingType:
        pass


class CanOutputProtocol(ABC):
    @abstractmethod
    def _to_event_output_type(self) -> OutputSetType:
        pass


class ObservableProtocol(CanInputProtocol):
    pass



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING
from .resolver import AssetResolver

if TYPE_CHECKING:
    from instaui.internal.backend.base import AssetBackendMixin


class BackendAssetResolver(AssetResolver):
    def __init__(self, backend: AssetBackendMixin):
        self.backend = backend

    def resolve_asset(self, path: Path) -> str:
        return self.backend.register_asset(path)



from typing import cast
from instaui.constants.ui import INSTAUI_JS_MODULE_NAME, USE_LANGUAGE_METHOD
from instaui.debug.api_boundary import user_api
from instaui.internal.ui.custom_var import CustomVar


@user_api
def use_language() -> str:
    """
    This function returns the current application's language setting, making it convenient to provide a unified language configuration for surrounding frameworks.
    """

    return cast(
        str,
        CustomVar(INSTAUI_JS_MODULE_NAME, USE_LANGUAGE_METHOD),
    )



from pathlib import Path
from typing import Union
import inspect


def get_caller_path(up: int = 2) -> Path:
    """
    Get the file path of the caller at the specified call stack level.

    Args:
        up: The number of call stack levels to go up. Default is 1 (direct caller).

    Returns:
        Path: The resolved file path of the caller.

    Raises:
        RuntimeError: If the call stack is too shallow.
    """
    return _get_caller_file(up)


def resolve_relative_path(path_or_str: Union[str, Path], up: int = 1) -> Path:
    if isinstance(path_or_str, Path):
        return path_or_str.resolve()

    caller_file = _get_caller_file(up + 1)
    return (caller_file.parent / path_or_str).resolve()


def _get_caller_file(up: int) -> Path:
    """
    Return the resolved Path of the caller file `up` levels above.
    """
    frame = inspect.currentframe()
    try:
        # +1 to skip this helper itself
        for _ in range(up + 1):
            if frame is None:
                raise RuntimeError("Frame stack too shallow")
            frame = frame.f_back

        if frame is None:
            raise RuntimeError("Frame stack too shallow")

        return Path(inspect.getfile(frame)).resolve()
    finally:
        del frame



import asyncio
import threading
from collections import defaultdict
from typing import Callable, Awaitable


class PageCache:
    def __init__(self) -> None:
        self._cache: dict[str, str] = {}

        self._async_locks: dict[str, asyncio.Lock] = defaultdict(asyncio.Lock)
        self._sync_locks: dict[str, threading.Lock] = defaultdict(threading.Lock)

    # ---------- async ----------
    async def get_or_render_async(
        self,
        *,
        key: str,
        cache: bool,
        render: Callable[[], Awaitable[str]],
    ) -> str:
        if not cache:
            return await render()

        # fast path
        html = self._cache.get(key)
        if html is not None:
            return html

        lock = self._async_locks[key]
        async with lock:
            # double check
            html = self._cache.get(key)
            if html is not None:
                return html

            html = await render()
            self._cache[key] = html
            return html

    # ---------- sync ----------
    def get_or_render_sync(
        self,
        *,
        key: str,
        cache: bool,
        render: Callable[[], str],
    ) -> str:
        if not cache:
            return render()

        # fast path
        html = self._cache.get(key)
        if html is not None:
            return html

        lock = self._sync_locks[key]
        with lock:
            # double check
            html = self._cache.get(key)
            if html is not None:
                return html

            html = render()
            self._cache[key] = html
            return html

    def warmup(self, key: str, html: str):
        self._cache[key] = html



from instaui.internal.pipeline.ast_build.base import RenderBuilderProtocol
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder

from instaui.internal.ui.vif import VIf
from instaui.internal.ast import core as ast_core
from instaui.internal.pipeline.ast_build.context import AstBuilderContext


class VIfParser:
    def __init__(
        self, ctx: AstBuilderContext, render_builder: RenderBuilderProtocol
    ) -> None:
        self.ctx = ctx
        self._expr_builder = ExprBuilder(ctx)
        self._render_builder = render_builder

    def parse(self, vif: VIf) -> ast_core.VIf:
        return ast_core.VIf(
            self._expr_builder.build(vif._on),
            children=[self._render_builder.build(r) for r in vif._renderables],
        )



from typing import Callable, Hashable, Mapping, Optional
from instaui.internal.backend.common import handler as _handler
from instaui.protocol.codec.type_adapter import TypeAdapterProtocol


_watch_handlers: dict[Hashable, _handler.HandlerInfo] = {}


def register_watch_handler(
    key: str,
    handler: Callable,
    outputs_binding_count: int,
    custom_type_adapter_map: Optional[Mapping[int, TypeAdapterProtocol]] = None,
):
    if key in _watch_handlers:
        return

    _watch_handlers[key] = _handler.HandlerInfo.from_handler(
        handler, outputs_binding_count, custom_type_adapter_map
    )


def get_watch_handler(key: str):
    return _watch_handlers.get(key)



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING, Optional
from instaui.systems.dataclass_system import dataclass

if TYPE_CHECKING:
    from .script_assets import ScriptTag
    from .plugin import PluginDependencyInfo
    from instaui.internal.assets.style_assets import StyleTag


@dataclass(frozen=True)
class AssetsSnapshot:
    js_links: list[str | Path]
    css_links: list[str | Path]
    style_tags: list[StyleTag]
    script_tags: list[ScriptTag]
    import_maps: dict[str, str | Path]
    favicon: Optional[str | Path]
    plugins: set[PluginDependencyInfo]
    component_dependencies: list[ComponentDependencySnapshot]


@dataclass(frozen=True)
class ComponentDependencySnapshot:
    tag: str
    esm: Path



from functools import wraps
from typing import Callable
from instaui.internal.ui.page_once import run_page_once


def page_once(
    func: Callable | None = None,
    *,
    key: str | None = None,
    namespace: str | None = None,
):
    """
    Executes a function at most once during a single page construction lifecycle.

    This decorator is intended to be used inside component constructors or page
    functions to register logic that should run only once per page render, even if
    the surrounding component is instantiated multiple times.

    Typical use cases include injecting JavaScript or CSS assets, registering global
    event handlers, or performing other one-time setup operations scoped to the
    current page.

    Args:
        func (Callable | None): The function to execute once. If None, the decorator
            returns a partially applied decorator for later use.
        key (str | None): An optional unique identifier used to deduplicate execution.
            Functions with the same key will be executed only once per page lifecycle.
            If not provided, a key will be automatically derived from the function's
            qualified name.
        namespace (str | None): Optional namespace used to isolate execution keys between
            different libraries or plugins. It is strongly recommended for plugin authors
            to set this value to their package or plugin name to avoid collisions.


    Example:
    .. code-block:: python
        from instaui import custom

        class CustomComponent(custom.element):
            def __init__(self):
                super().__init__()

                @custom.page_once
                def inject_runtime():
                    self.add_js("console.log('loaded')")


        # With explicit key for cross-component deduplication
        @custom.page_once(key="shiki_runtime")
        def install_shiki():
            load_shiki_assets()

    Notes:
        - The function is executed immediately at decoration time.
        - The execution is scoped to the current page context and isolated between
          different page renders and user sessions.
        - Calling this decorator outside of a page context will raise an error.
    """
    if func is None:
        return lambda f: page_once(f, key=key, namespace=namespace)

    @wraps(func)
    def wrapper(*args, **kwargs):
        return run_page_once(
            lambda: func(*args, **kwargs),
            key=key,
            namespace=namespace,
        )

    # execute immediately
    wrapper()

    return wrapper



_DEV_MODE = False
_FROZEN = False


def set_dev(value: bool) -> None:
    global _DEV_MODE, _FROZEN
    if _FROZEN:
        raise RuntimeError(
            "instaui dev mode is already frozen; "
            "import instaui.dev must be called before importing ui"
        )
    _DEV_MODE = bool(value)


def is_dev() -> bool:
    return _DEV_MODE


def freeze() -> None:
    """
    Called by subsystems that rely on dev/prod mode
    to prevent late mutation.
    """
    global _FROZEN
    _FROZEN = True



from __future__ import annotations
from typing import Any, Optional, Literal, Union
from instaui.internal.ui.element import Element
from ._responsive_type._common import (
    TMaybeResponsive,
    TLevel_1_9,
)
from ._responsive_type._typography import (
    TWeightEnum,
    TTextWrapEnum,
    TTrimEnum,
    TAlignEnum,
)


class Text(Element):
    def __init__(
        self,
        text: Optional[Any] = None,
        *,
        as_: Optional[Literal["span", "div", "label", "p", "pre"]] = None,
        as_child: Optional[bool] = None,
        size: Optional[TMaybeResponsive[TLevel_1_9]] = None,
        weight: Optional[TMaybeResponsive[Union[TWeightEnum, str]]] = None,
        align: Optional[TMaybeResponsive[Union[TAlignEnum, str]]] = None,
        trim: Optional[Union[TTrimEnum, str]] = None,
        truncate: Optional[bool] = None,
        text_wrap: Optional[Union[TTextWrapEnum, str]] = None,
    ):
        """
        Creates a text element with customizable styling and typography options.

        Args:
            text (Optional[Any]]: The text content to display.
            as_ (Optional[Literal["span", "div", "label", "p", "pre"]]]):
                HTML element type to render as. Defaults to appropriate semantic element.
            as_child (Optional[bool]): Whether to merge props with parent component.
            size (Optional[TMaybeResponsive[TLevel_1_9]]): Text size level from 1 to 9.
            weight (Optional[TMaybeResponsive[Union[TWeightEnum, str]]): Font weight.
            align (Optional[TMaybeResponsive[Union[TAlignEnum, str]]): Text alignment.
            trim (Optional[Union[TTrimEnum, str]]): Whitespace trimming behavior.
            truncate (Optional[bool]): Whether to truncate overflowing text.
            text_wrap (Optional[Union[TTextWrapEnum, str]]): Text wrapping behavior.

        Example:
        .. code-block:: python
            # Basic text element
            ui.text("ui.text")

            # Text with specific HTML element and styling
            ui.text(
                "Styled text",
                as_="div",
                size=3,
                weight="bold",
                align="center"
            )
        """
        super().__init__("ui-text")

        self.props(
            {
                "innerText": text,
                "as": as_,
                "as_child": as_child,
                "size": size,
                "weight": weight,
                "text_align": align,
                "trim": trim,
                "truncate": truncate,
                "text_wrap": text_wrap,
            }
        )



from __future__ import annotations
from typing import TYPE_CHECKING
from instaui.internal.codegen.context.variable_names import VariableNameContext


if TYPE_CHECKING:
    from .root_ctx import CodegenRootContext


class FileCodegenContext:
    def __init__(self, root: CodegenRootContext):
        self.root = root

        # file-scoped
        self.names = VariableNameContext()
        self.imports = root.components.import_table_factory()



from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, ClassVar, Optional

from instaui.constants.runtime import RuntimeMode

if TYPE_CHECKING:
    from instaui.internal.ui._scope import Scope
    from instaui.internal.ui.pending_scope import PendingScope
    from instaui.internal.ui.container import Container
    from instaui.internal.ui.layout import Layout


class App:
    _default_app: ClassVar[Optional[App]] = None

    def __init__(
        self,
        mode: RuntimeMode,
        *,
        debug: bool,
        meta: Optional[dict] = None,
    ) -> None:
        super().__init__()
        self._scoped_style_id_counter = 0
        self._scope_id_counter = 0
        self._container_stack: list[Container] = []
        self._scope_stack: list[Scope] = []
        self._pending_scope_stack: list[PendingScope] = []
        self._created_scopes: list[Scope] = []
        self._mode: RuntimeMode = mode
        self._debug_mode: bool = debug
        self.meta = meta

        self._used_icons: set[str] = set()
        self._used_icons_set: set[str] = set()
        self.layouts: list[Layout] = []
        self._page_once_registry: set[str] = set()

    @property
    def created_scopes(self):
        return self._created_scopes

    def next_scope_id(self) -> int:
        self._scope_id_counter += 1
        return self._scope_id_counter

    def append_layout(self, layout: Layout) -> None:
        self.layouts.append(layout)

    def push_scope(self, scope: Scope):
        self._scope_stack.append(scope)
        self._created_scopes.append(scope)

    def pop_scope(self) -> Scope:
        return self._scope_stack.pop()

    def push_pending_scope(self, pending_scope: PendingScope):
        self._pending_scope_stack.append(pending_scope)

    def pop_pending_scope(self) -> PendingScope:
        return self._pending_scope_stack.pop()

    def has_pending_scope(self) -> bool:
        return len(self._pending_scope_stack) > 0

    def top_pending_scope(self):
        return self._pending_scope_stack[-1]

    def push_container(self, container: Container):
        self._container_stack.append(container)

    def pop_container(self) -> Container:
        return self._container_stack.pop()

    def get_current_scope(self) -> Scope:
        assert len(self._scope_stack) > 0, "Scope stack is empty"
        return self._scope_stack[-1]

    def get_current_container(self) -> Container:
        assert len(self._container_stack) > 0, "Container stack is empty"
        return self._container_stack[-1]

    def collect_icon(self, icon_name: str):
        self._used_icons.add(icon_name)

    def collect_icon_set(self, icon_set_name: str):
        self._used_icons_set.add(icon_set_name)

    @contextmanager
    def _mark_router_base_scope(self, scope: Scope):
        self._router_base_scope = scope
        yield
        self._router_base_scope = None

    @property
    def router_base_scope(self):
        return self._router_base_scope

    @property
    def mode(self) -> RuntimeMode:
        return self._mode

    @property
    def debug_mode(self) -> bool:
        return self._debug_mode

    def gen_scoped_style_group_id(self):
        gid = f"scoped-style-{self._scoped_style_id_counter}"
        self._scoped_style_id_counter += 1
        return gid


class DefaultApp(App):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(DefaultApp, cls).__new__(cls)
        return cls._instance

    def new_scope(self):
        raise ValueError("Operations are not allowed outside of ui.page")


App._default_app = DefaultApp(mode=RuntimeMode.WEB, debug=False)



from typing import Callable
from instaui.internal.ui.app_context import get_app


def run_page_once(func: Callable[[], None], key=None, namespace: str | None = None):
    app = get_app()
    if app is None:
        raise RuntimeError("custom.page_once must be called inside a page function")

    once_key = make_once_key(func, key, namespace)

    if once_key in app._page_once_registry:
        return False

    app._page_once_registry.add(once_key)
    func()
    return True


def make_once_key(
    func: Callable[[], None], key: str | None = None, namespace: str | None = None
) -> str:
    if namespace is None:
        namespace = func.__module__

    if key is None:
        key = func.__qualname__

    return f"{namespace}:{key}"



from instaui.internal import import_presets
from instaui.internal.ast.core import (
    ContentRender,
    Element,
    MatchRender,
    Render,
    VFor,
    ComponentRefRender,
    VIf,
)
from instaui.internal.codegen.component_ref_renderer import ComponentRefRenderer
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.element_renderer import ElementRenderer
from instaui.internal.codegen.expr_codegen import ExpressionCodegen
from instaui.internal.codegen.match_renderer import MatchRenderer
from instaui.internal.codegen.vfor_renderer import VForRenderer
from instaui.internal.codegen.vif_renderer import VIfRenderer


class RenderCodegen:
    def __init__(self, ctx: FileCodegenContext) -> None:
        self.ctx = ctx
        self.expr = ExpressionCodegen(ctx)
        self._element_renderer = ElementRenderer(ctx, render_emitter=self)
        self._vfor_renderer = VForRenderer(ctx, render_emitter=self)
        self._vif_renderer = VIfRenderer(ctx, render_emitter=self)
        self._match_renderer = MatchRenderer(ctx, render_emitter=self)
        self._component_ref_renderer = ComponentRefRenderer(ctx)

    def emit_renders(self, renders: list[Render]) -> str:
        if not renders:
            return "null"

        return f"[{', '.join(self.emit_render(render) for render in renders)}]"

    def emit_render(self, render: Render) -> str:
        method = getattr(self, f"emit_{type(render).__name__}", None)
        if not method:
            raise NotImplementedError(type(render))
        return method(render)

    def emit_Element(self, render: Element) -> str:
        return self._element_renderer.emit(render)

    def emit_VFor(self, render: VFor) -> str:
        return self._vfor_renderer.emit(render)

    def emit_VIf(self, render: VIf) -> str:
        return self._vif_renderer.emit(render)

    def emit_MatchRender(self, render: MatchRender) -> str:
        return self._match_renderer.emit(render)

    def emit_ComponentRefRender(self, render: ComponentRefRender) -> str:
        return self._component_ref_renderer.emit(render.ref)

    def emit_ContentRender(self, content: ContentRender) -> str:
        method = self.ctx.imports.use_from_preset(
            import_presets.Instaui.render_content()
        )
        return f"{method}({self.expr.emit(content.content)})"



from .registry import route_registry
from .model import PageInfo


def page(path: str, *, cache: bool = True):
    """
    Decorator for page registration.

    Args:
        path (str): Path of the page.
        cache (bool, optional): Whether to cache the page or not. Defaults to True.
            Note: When path contains parameters (e.g. '/{name}'), cache should be set to False
            as parameterized paths typically serve dynamic content that shouldn't be cached.

    Example:
    .. code-block:: python
        from instaui import ui

        @ui.page('/')
        def home():
            ui.text('Hello, world!')

        @ui.page('/about')
        def about():
            ui.text('About us')

        @ui.page('/{name}', cache=False)
        def user():
            name = ui.param('name')
            ui.text(name)
    """

    def decorator(func):
        route_registry.add_route(path, PageInfo(func, cache))
        return func

    return decorator



from abc import ABC, abstractmethod


class RouteSyntaxAdapter(ABC):
    @abstractmethod
    def normalize(self, path: str) -> str:
        """
        Convert framework-supported route DSL
        into frontend-parseable route pattern.

        Example:
            "/users/{id}" -> "/users/:id"
        """
        pass



from pathlib import Path
from typing import Final


_PROJECT_DIR: Final = Path(__file__).parent
STATIC_DIR: Final = _PROJECT_DIR.joinpath("static")
COMPILED_DIR: Final = STATIC_DIR.joinpath("compiled")

FAVICON_PATH: Final = STATIC_DIR.joinpath("insta-ui.ico")

# compiled files
APP_ES_JS_PATH: Final = COMPILED_DIR.joinpath("insta-ui.esm-browser.prod.js")
APP_ES_JS_MAP_PATH: Final = COMPILED_DIR.joinpath("insta-ui.js.map")
APP_CSS_PATH: Final = COMPILED_DIR.joinpath("insta-ui.css")
VUE_ES_JS_PATH: Final = COMPILED_DIR.joinpath("vue.esm-browser.prod.js")
